/***************************************************************************
 *   Copyright (C) 2007 by Gunnar Kjemphol   *
 *   gunnar@kjemphol.no   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <cstdlib>

#include "astartest.h"
#include "stlastar.h"
#include "MapSearchNode.h"
#include "WaypointGridMapping.h"

const int WORLD_GRID_X = 5;
const int WORLD_GRID_Y = 10;

int worldMap[WORLD_GRID_X * WORLD_GRID_Y];

using namespace std;

int InitMap ( int size_x, int size_y ){
  int i;
  for(i = 0; i < (size_x * size_y); i++){
    worldMap[i] = 1; 
  }
  
  return 0;
}

int SetMapWeight( int x, int y, int weight ){
  if(weight < 1)
    weight = 1;
  else if(weight > 9)
    weight = 9;
  if( x < 0 ||
      x >= WORLD_GRID_X ||
      y < 0 ||
      y >= WORLD_GRID_Y
    )
  {
    return -1;
  }
  else {
    worldMap[(y*WORLD_GRID_X) + x ] = weight;
  }
  
  return 0;
}
/*
int PrintWorldMap(){
  int x,y;
  cout << "worldMap " << endl;
  for(y = 0; y < WORLD_GRID_Y; y++) {
   for(x = 0; x < WORLD_GRID_X; x++){
     cout << worldMap[(y*WORLD_GRID_X) + x ];
   }
   cout << endl;
 }
 cout << endl;

}
*/
int PrintWorldMap(){
  int x,y;
  cout << "worldMap " << endl;
  for(y = WORLD_GRID_Y-1; y >= 0; y--) {
    for(x = WORLD_GRID_X-1; x >= 0; x--){
      cout << worldMap[(y*WORLD_GRID_X) + x ];
    }
    cout << endl;
  }
  cout << endl;
}


int GetMap( int x, int y ){

  if( x < 0 ||
      x >= WORLD_GRID_X ||
      y < 0 ||
      y >= WORLD_GRID_Y
    )
  {
    return 9;	 
  }

  return worldMap[(y*WORLD_GRID_X)+x];
}


int main(int argc, char *argv[]) {
  cout << "A* Test program" << endl;
  PrintWorldMap();
  InitMap(WORLD_GRID_X, WORLD_GRID_Y);
  PrintWorldMap();



  MapSearchNode nodeStart;
  nodeStart.x = 0;
  nodeStart.y = 0;

  MapSearchNode nodeEnd;
  nodeEnd.x = 3;
  nodeEnd.y = 6;

  /**
   * Set weights in map
   */
  /*
  SetMapWeight(2,3,5);
  SetMapWeight(1,3,5);
  SetMapWeight(3,3,5);
  SetMapWeight(2,4,5);
  SetMapWeight(2,6,5);
  SetMapWeight(3,4,5);
  SetMapWeight(3,5,5);
  */
  
  SetMapWeight(2,5,5);
  PrintWorldMap();

  /**
   * Set start and goal states
   */
  AStarSearch<MapSearchNode> astarsearch;
  astarsearch.SetStartAndGoalStates( nodeStart, nodeEnd );

  unsigned int SearchState;
  unsigned int SearchSteps = 0;

  do
  {
    SearchState = astarsearch.SearchStep();

    SearchSteps++;

#if DEBUG_LISTS

    cout << "Steps:" << SearchSteps << "\n";

    int len = 0;

    cout << "Open:\n";
    MapSearchNode *p = astarsearch.GetOpenListStart();
    while( p ){
      len++;
#if !DEBUG_LIST_LENGTHS_ONLY
      ((MapSearchNode *)p)->PrintNodeInfo();
#endif
      p = astarsearch.GetOpenListNext();
    }

    cout << "Open list has " << len << " nodes\n";

    len = 0;

    cout << "Closed:\n";
    p = astarsearch.GetClosedListStart();
    while( p )
    {
      len++;
#if !DEBUG_LIST_LENGTHS_ONLY			
      p->PrintNodeInfo();
#endif			
      p = astarsearch.GetClosedListNext();
    }

    cout << "Closed list has " << len << " nodes\n";
#endif

  }while( SearchState == AStarSearch<MapSearchNode>::SEARCH_STATE_SEARCHING ); //do 


  if( SearchState == AStarSearch<MapSearchNode>::SEARCH_STATE_SUCCEEDED )
  {
    cout << "Search found goal state\n";

    MapSearchNode *node = astarsearch.GetSolutionStart();

#if DISPLAY_SOLUTION
    cout << "Displaying solution\n";
#endif
    int steps = 0;

    node->PrintNodeInfo();
    for( ;; )
    {
      node = astarsearch.GetSolutionNext();

      if( !node )
      {
	break;
      }

      node->PrintNodeInfo();
      steps ++;

    };

    cout << "Solution steps " << steps << endl;

  /**
   * Once you're done with the solution you can free the nodes up
   */
  astarsearch.FreeSolutionNodes();

  }
  else if( SearchState == AStarSearch<MapSearchNode>::SEARCH_STATE_FAILED ){
    cout << "Search terminated. Did not find goal state\n";
  }
  /**
   * Display the number of loops the search went through
   */
  cout << "SearchSteps : " << SearchSteps << "\n";


  astarsearch.EnsureMemoryFreed();


  return EXIT_SUCCESS;
}
