/*
** ** ** ** ** **  Begin header  ** ** ** ** ** ** **
**
** Project : Byggelab i faget TTK4155 Industrielle Datasystemers Konstruksjon 2005
** Filename : servo_pwm_driver.c
** Version : 1
** Date : 07.11.2005
**
** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** **
**
** Copyright (c) 2005, Kristian M. Knausgrd, Gunnar Kjemphol
** All rights reserved.
**
** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** ** **

VERSION LOG:
-----------
Version: 1
Date: 07.11.2005
Authors: Kristian M. Knausgrd, Gunnar Kjemphol
Coauthors: <none>
Description : Original version.
Target platform: ATMega128

FILE FUNCTIONALITY DESCRIPTION:
-----------
Servo driver for 1 standard PWM servo using timers on ATMega128.

** ** ** ** ** ** ** End header ** ** ** ** ** ** ** **
*/


#include <avr/io.h>
#include "servo_pwm_driver.h"

static int ocr_bias;

#define OCR_CENTER 832


/**
* Start timer in pwm mode, and set servo positioned at neutral position (0 degrees).
*/
void servo_init( void ) {

	ocr_bias = 0;

	// Set PWM output pin as output
	DDRB |=(1<<PB5);

	// Set up PWM 16 bit timer on ATMEGA128
	TCCR1A =(1<<COM1A1);

	// Using 8 for 8MHz clock. (clockprescale = (f_osc*0.02/1024))
	TCCR1B = (1<<WGM13)|(1<<CS11)|(1<<ICNC1);
	// Top value for counter (turns here..) This gives us the pwm signal period. Duty cycle is set by servo_set_position.
	ICR1 = 10016;

	// Start PWM with correct neutral PWM signal on servo
	servo_set_position_deg(0);
}



/**
* Set servo position in degrees (-90 to 90),
*/
void servo_set_position_deg(int angle_deg) {

	// Ensure input value is in valid range. Bad output PWM signal to servo may damage it.
	if(angle_deg<=90 && angle_deg>=-90) {
		volatile int ocr_value;	//TODO: remove volatile??
		ocr_value=(OCR_CENTER + (int)(((float)angle_deg/90)*500) + ocr_bias);
		OCR1A = ocr_value;
	}
}


/**
* Set servo position in radians (-pi to pi).
*/
void servo_set_position_rad(double angle_rad) {

}

void set_servo_center(void){
	ocr_bias = OCR1A - OCR_CENTER;
}
