/**
 *  Eurobot 2007
 *  BatterySort.cpp
 *  Gunnar Kjemphol <gunnar@kjemphol.no>
 */

#include "BatterySort.h"


BatterySort::BatterySort ( PlansysObservation *s , int theTeamColor ){
  teamColor = theTeamColor;
  /**
   * Give the Agent a priority in the constructor [0-9]
   * if the agentPriority is -1 the agent is inactive
   */
  agentPriority = 5;
  inAction = INACTIVE;
  activeStrategy = INACTIVE;
  batteryRelease = INACTIVE;
  batteryReleased = FALSE;
  lastWaypointReached = FALSE;
  waitForBatteryReleased = 0;
  _subject = s;
  _subject->Attach(this);
    /**
     * Init the battery bin in midt position of three
     */
    battery_bin_ai_waypoint.x_l = 1500;
    battery_bin_ai_waypoint.y_l = 1400 - 100; // 1400 - radius bin
    battery_bin_ai_waypoint.psi_l = M_PI;
    battery_bin_ai_waypoint.circle_of_acceptance = 30;
    
    clear_controll_struct(&local_controll_struct);

}

BatterySort::~BatterySort (){
  _subject->Detach(this);

}

int BatterySort::AgentGlobalPriority(){
  return agentPriority; 
}

int BatterySort::SetAgentActive( int active ){
  activeStrategy = active;
  if ( active == INACTIVE)
    inAction = INACTIVE;
}

int BatterySort::GetPriority(){
  int priority = -100;
  float cv_x, cv_y;
  ai_waypoint local_ai_waypoint;
  /**
   * Lock for battery
   */
#ifdef DEBUG_BATTERY_SORT
  cout << " " << endl;
  cout << "BATTERYSORT: Strategy Active = " << activeStrategy << endl;
  cout << " " << endl;
  cout << "inAction " << inAction << endl;
  cout << "batteryRelease " << batteryRelease << endl;
  cout << "batteryReleased " << batteryReleased << endl;
  cout << " lastWaypointReached " << lastWaypointReached << endl;
  cout << " \n\n " << endl;
#endif
  if( inAction == INACTIVE ) {
    cout << "inAction == INACTIVE << " << endl;
    //printf("my_cv_observation.type %d \n", my_cv_observation.type);
    //cout << "my_cv_observation.type " << my_cv_observation.type << endl;
    if( my_cv_observation.type == BLUE_BATTERY ){
      //cout << "my_cv_observation.type == BLUE_BATTERY " << endl;
      //cout << "Team Color =  " << teamColor << endl;
      if( teamColor == TEAM_COLOR_BLUE ) {
	cout << "cv_observation BLUE_BATTERY  and TEAM_COLOR_BLUE " << endl; 
	//cout << "teamColor == TEAM_COLOR_BLUE " << endl;
	/**
	 * Set waypoints to get the battery. 
	 * Set full priority.
	 * Use the A* star to go to the bin
	 * The bin position is in the battery_bin_ai_waypiont struct
	 */
	
	cv_x = (float)(my_cv_observation.x);
	cv_y = (float)(my_cv_observation.y);

	local_ai_waypoint.psi_l = atan2f(cv_x, cv_y);
	local_ai_waypoint.x_l = my_cv_observation.x;
	local_ai_waypoint.y_l = my_cv_observation.y;
	
	/**
	 * The position of the battery, the battery will be picked up.
	 */	
	local_controll_struct._ai_waypoint_rel_robot.push_back(local_ai_waypoint);
	
	/**
	 * Use A* her to get the waypoints to the battery bin
	 */
	
	/**
	 * Go to the table center, 400 mm away from battery bin;
	 */
	local_ai_waypoint.x_l = battery_bin_ai_waypoint.x_l;
	local_ai_waypoint.y_l = battery_bin_ai_waypoint.y_l - 400;
	local_ai_waypoint.psi_l = battery_bin_ai_waypoint.psi_l;
	local_ai_waypoint.circle_of_acceptance = 50;
	local_controll_struct._ai_waypoint.push_back(local_ai_waypoint);
	
	/**
	 * Set inn the battery bin waypoint
	 */
	local_controll_struct._ai_waypoint.push_back(battery_bin_ai_waypoint);
        /**
	 * Slett at man har sett batteri
         */
	my_cv_observation.type = NOTHING;
	inAction = ACTIVE;
	
	priority = 120;
      }
    }
    else if ( my_cv_observation.type == RED_BATTERY ) {
      if( teamColor == TEAM_COLOR_RED ) {
	/**
	 * Set waypoints to get the battery. 
	 * Set full priority.
	 * Use the A* star to go to the bin
	 * The bin position is in the battery_bin_ai_waypiont struct
	 */
	
	cv_x = (float)(my_cv_observation.x);
	cv_y = (float)(my_cv_observation.y);

	local_ai_waypoint.psi_l = atan2f(cv_x, cv_y);
	local_ai_waypoint.x_l = my_cv_observation.x;
	local_ai_waypoint.y_l = my_cv_observation.y;
	
	/**
	 * The position of the battery, the battery will be picked up.
	 */	
	cout << "BATTERY_SORT _ai_waypoint_rel_robot.push_back" << endl;
	local_controll_struct._ai_waypoint_rel_robot.push_back(local_ai_waypoint);
	
	/**
	 * Use A* her to get the waypoints to the battery bin
	 */
	
	/**
	 * Set inn the battery bin waypoint
	 */
	cout << "BATTERY_SORT _ai_waypoint.push_back" << endl;
	local_controll_struct._ai_waypoint.push_back(battery_bin_ai_waypoint);
        /**
	 * Slett at man har sett batteri
	 */
	my_cv_observation.type = NOTHING;
	inAction = ACTIVE;
	priority = 120;	
      }
    }
      /**
       * Proceed if not activeStrategy and no battery
       * Return -100
       */
    else if ( my_cv_observation.type == NOTHING ) {
      cout << "BatterySort: my_cv_observation.type == NOTHING " << endl;
      cout << "BatterySort: inAction = " << inAction << endl;
      //inAction = INACTIVE;
      return -200;
	
    }
    else {
      cout << "BatterySort: my_cv_observation.type = " << my_cv_observation.type<< endl;
      cout << "BatterySort: inAction = " << inAction << endl;
      //inAction = INACTIVE;
      return -200;
	
    }
  } //if( inAction == INACTIVE ) {


  cout << "activeStrategy " << activeStrategy << endl; 
  if( (inAction == ACTIVE) && (batteryRelease == INACTIVE)) {
#ifdef DEBUG_BATTERY_SORT
    cout << "\n\n\n(inAction == ACTIVE) && (batteryRelease == INACTIVE)) \n\n" << endl;
#endif
    /**
     * Check if the robot is in battery bin position
     */
    cout << "lastWaypointReached " << lastWaypointReached << endl;
    if( lastWaypointReached == TRUE ){
      cout << "lastWaypointReached == TRUE" << lastWaypointReached << endl;

      /**
       * Set controlstruct command to release battery when in position
       */
      local_controll_struct.command.push_back( MINOR_CMD_RELEASE_BATTERY );
      lastWaypointReached = FALSE;
      /**
       * Stop the robot and wait for the robot to release te battery
       */
      local_controll_struct.setMaxSpeed = true;
      local_controll_struct.maxSpeed_mps = 0.0;
      /**
       * Very high priority
       */

      priority = 1000;

     // inAction = INACTIVE;
#ifdef DEBUG_BATTERY_SORT
      cout << "\n\n\n\n\n" << endl;
      cout << "batteryRelease = ACTIVE;" << batteryRelease << endl;
      cout << "batteryRelease = ACTIVE;" << batteryRelease << endl;
      cout << "batteryRelease = ACTIVE;" << batteryRelease << endl;
      cout << "batteryRelease = ACTIVE;" << batteryRelease << endl;
      cout << "\n\n\n\n\n" << endl;
#endif
      batteryRelease = ACTIVE;

    }
  }
  else if( (inAction == ACTIVE) && (batteryRelease == ACTIVE)) {
/**
 * 
 *     Finn en løsning som venter en stund på at batteri skal 
 *    senkes ned før den rygger.
 */
//cout << "inAction " << inAction << endl;
//cout << "batteryReleased " << batteryReleased << endl;

    if( batteryReleased == TRUE ){
      cout << "if( batteryReleased == TRUE ) \n\n\n\n\n"<< endl;
      cout << "if( batteryReleased == TRUE )\n\n\n\n\n\n" << endl;

      local_controll_struct.clearWaypoints = true;
      local_controll_struct.clearTheActiveStrategy = true;
      local_controll_struct.setMaxSpeed = true;
      local_controll_struct.maxSpeed_mps = MAX_SPEED_MPS;
      /**
       * Go away form battery bin
       */
      local_ai_waypoint.x_l = -300;
      local_ai_waypoint.y_l = 0;
      local_ai_waypoint.psi_l = M_PI;
      local_ai_waypoint.circle_of_acceptance = 100;
      local_controll_struct._ai_waypoint_rel_robot.push_back(local_ai_waypoint);
      /**
       * Turn against bottle and box bins
       */
      local_ai_waypoint.x_l = 0;
      local_ai_waypoint.y_l = 0;
      if(teamColor == TEAM_COLOR_BLUE)
        local_ai_waypoint.psi_l = M_PI/8;
      else
	local_ai_waypoint.psi_l = (7 * M_PI)/8;
      local_ai_waypoint.circle_of_acceptance = 100;
      local_controll_struct._ai_waypoint_rel_robot.push_back(local_ai_waypoint);

      batteryReleased = FALSE;
      batteryRelease = INACTIVE;
      waitForBatteryReleased = 0;
    }
    else {
    waitForBatteryReleased = waitForBatteryReleased + 1;
#ifdef DEBUG_BATTERY_SORT
    cout << " " << endl;
    cout << " " << endl;
    cout << "before waitForBatteryReleased = " << waitForBatteryReleased << endl;
    cout << " " << endl;
    cout << " " << endl;
#endif
    }
    if((waitForBatteryReleased < 15)) {
      waitForBatteryReleased = waitForBatteryReleased + 1;
#ifdef DEBUG_BATTERY_SORT
      cout << " " << endl;
      cout << " " << endl;
      cout << "waitForBatteryReleased = " << waitForBatteryReleased << endl;
      cout << " " << endl;
      cout << " " << endl;
#endif
    }
    else {

      cout << " " << endl;
      cout << " " << endl;
      cout << "BatterySort: ERROR Handling " << endl;
      cout << "BatteryRelease timeout set BatterySort inactive " << endl;
      cout << " " << endl;
      cout << " " << endl;
      agentPriority = -1;
      waitForBatteryReleased = 1000;
      
    }
  }

  cout << "inAction " << inAction << endl;
  cout << "batteryRelease " << batteryRelease << endl;
  cout << "batteryReleased " << batteryReleased << endl;
  cout << " lastWaypointReached " << lastWaypointReached << endl;


  return priority;
}

int BatterySort::GetAction( controll_struct *ai_controll_struct ){
  
  cout << " ACTIVE STRATEGY BATTERYSORT " << endl;
  memcpy((void *)(ai_controll_struct),( void * )&local_controll_struct, sizeof(controll_struct));
  /**
   * Clear contol_struct
   */
  clear_controll_struct( &local_controll_struct );


  return 0;
}

void BatterySort::Update( Subject *theChangedSubject, int typeOfObservation){
  if( theChangedSubject == _subject ){
    if( typeOfObservation == MINOR_PROXY_MESSAGE ){
      my_proxy_message = _subject->GetProxyMessage();
      if( my_proxy_message.teamColor > 0 )
	teamColor = my_proxy_message.teamColor;

    }
    else if ( typeOfObservation == MINOR_NAVSYS_REACHED_WAYPOINT ) {
#ifdef DEBUG_BATTERY_SORT
      cout << "\n\n BATTERY MINOR_NAVSYS_REACHED_WAYPOINT \n\n" << endl;
#endif
      ai_waypoint last_ai_waypoint_3d_reached;
      last_ai_waypoint_3d_reached = _subject->GetWaypointReachedObservation();
      if((last_ai_waypoint_3d_reached.x_l == battery_bin_ai_waypoint.x_l) && (last_ai_waypoint_3d_reached.y_l == battery_bin_ai_waypoint.y_l))
        lastWaypointReached = TRUE;
#ifdef DEBUG_BATTERY_SORT
      if(lastWaypointReached == TRUE)
        cout << "\n \n \n \n BatterySort BATTERY_BIN REACHED \n \n \n \n \n " << endl;
#endif
    }
    else if ( typeOfObservation == MINOR_CMD_BATTERY_RELEASED ) {
      batteryReleased = TRUE;
#ifdef DEBUG_BATTERY_SORT
      cout << "\n \n \n \n BatterySort: batteryReleased = TRUE; \n \n \n \n \n " << endl;
#endif
    }
    else if( typeOfObservation == MINOR_NAVSYS_POSITION ){
      ai_waypoint_3d = _subject->GetPosObservation();
//#ifdef DEBUG_OBS_PRINT
      cout << "Type of observation POSITION BatterySort" << endl;
      cout << "x_l = " << ai_waypoint_3d.x_l << endl;
      cout << "y_l = " << ai_waypoint_3d.y_l << endl;
      cout << "psi_l = " << ai_waypoint_3d.psi_l << endl;
//#endif
     }
     else if( typeOfObservation == MINOR_CV_OBSERVATION ){
        my_cv_observation = _subject->GetCvObservation();
#ifdef DEBUG_OBS_PRINT
        cout << "Type of observation CV BatterySort" << endl;
        cout << "my_cv_observation.x = " << my_cv_observation.x << endl; 
        cout << "my_cv_observation.y = " << my_cv_observation.y << endl;
#endif
	if( my_cv_observation.type == BATTERY_BIN ) {
	  /**
	   * Nead tuning
	   */
	  battery_bin_ai_waypoint.x_l = my_cv_observation.x;
	  battery_bin_ai_waypoint.y_l = my_cv_observation.y;

	}
      }
#ifdef DEBUG_OBS_PRINT
      cout << "Update BatterySort" << endl;
      cout << " " << endl;
#endif


  }
}

/**
 * The strategy is interrupted by anoter strategy
 * keep existing settings, or roll back to last setting
 * for retransmitting the commands or waypoints
 */
int BatterySort::StrategyInterrupted() {

  
  return 0;
}

int BatterySort::GetInterruptAction(interrupt_controll_struct *interrupt_controll) {
  
 return 0; 
}


