/**
 *  Eurobot 2007
 *  PlansysObservation.cpp
 *  Gunnar Kjemphol <gunnar@kjemphol.no>
 */

#include "BottleBoxSort.h"


BottleBoxSort::BottleBoxSort ( PlansysObservation *s , int theTeamColor ){
  teamColor = theTeamColor;
  /**
   * Give the Agent a priority in the constructor [0-9]
   * if the agentPriority is -1 the agent is inactive
   */
  agentPriority = 5;
  inAction = INACTIVE;
  activeStrategy = INACTIVE;
  bottleBoxRelease = INACTIVE;
  bottleBoxReleased = FALSE;
  lastWaypointReached = FALSE;
  waitForBottleBoxReleased = 0;
  typeOfPickUp = NOTHING;
  pickUpMessageRecieved = false;
  _subject = s;
  _subject->Attach(this);

  /**
   * Bottle and bin positions
   * Two for each teamColor
   */

  if( teamColor == TEAM_COLOR_RED ) {
    red_bottle_bin_ai_waypoint.x_l = 700;
    red_bottle_bin_ai_waypoint.y_l = 1600;
    red_bottle_bin_ai_waypoint.psi_l = ((7*M_PI)/8);
    red_bottle_bin_ai_waypoint.circle_of_acceptance = 50;

    red_box_bin_ai_waypoint.x_l = 500;
    red_box_bin_ai_waypoint.y_l = 1400;
    red_box_bin_ai_waypoint.psi_l = ((5*M_PI)/8);
    red_box_bin_ai_waypoint.circle_of_acceptance = 50;

  }
  else if ( teamColor == TEAM_COLOR_BLUE ) {
    blue_bottle_bin_ai_waypoint.x_l = 2300;
    blue_bottle_bin_ai_waypoint.y_l = 1600;
    blue_bottle_bin_ai_waypoint.psi_l = ((1*M_PI)/8);
    blue_bottle_bin_ai_waypoint.circle_of_acceptance = 50;

    blue_box_bin_ai_waypoint.x_l = 2500;
    blue_box_bin_ai_waypoint.y_l = 1400;
    blue_box_bin_ai_waypoint.psi_l = ((3*M_PI)/8);
    blue_box_bin_ai_waypoint.circle_of_acceptance = 50;

  }
  else {
    /**
     * Error team color, ask for team color
     */
    cout << "BottleBox: No teamColor " << endl;
  }

  clear_controll_struct(&local_controll_struct);

}



BottleBoxSort::~BottleBoxSort (){
  _subject->Detach(this);

}

int BottleBoxSort::AgentGlobalPriority(){
  return agentPriority; 
}

int BottleBoxSort::SetAgentActive( int active_r ){
  inAction = active_r;
  activeStrategy = ACTIVE;
  if( active_r == INACTIVE) {
    inAction = INACTIVE;
    activeStrategy = INACTIVE;
  }
}


int BottleBoxSort::GetPriority(){
  int priority = -100;
  float cv_x, cv_y;
  ai_waypoint local_ai_waypoint;
  /**
   * Lock for battery
   */
#ifdef DEBUG_BOTTLE_BOX_SORT
  cout << " " << endl;
  cout << "BOTTLE_BOX_SORT: Strategy Active = " << activeStrategy << endl;
  cout << " " << endl;
  cout << "BOTTLE_BOX_SORT: inAction " << inAction << endl;
  cout << "BOTTLE_BOX_SORT: bottleBoxRelease " << bottleBoxRelease << endl;
  cout << "BOTTLE_BOX_SORT: bottleBoxReleased " << bottleBoxReleased << endl;
  cout << "BOTTLE_BOX_SORT: lastWaypointReached " << lastWaypointReached << endl;
  cout << " \n\n " << endl;
#endif
  if( inAction == INACTIVE ) {
    cout << "BOTTLE_BOX_SORT: inAction == INACTIVE << " << endl;

    if( my_cv_observation.type == BOX ){
	/**
	 * Set waypoints to get the BOX
	 * Set full priority.
	 * Use the A* star to go to the bin
	 * The bin position is in the box_bin_ai_waypiont struct
	 */
	
	cv_x = (float)(my_cv_observation.x);
	cv_y = (float)(my_cv_observation.y);

	local_ai_waypoint.psi_l = atan2f(cv_y, cv_x);
	local_ai_waypoint.x_l = my_cv_observation.x + 50; // Add 50mm to be sure
	local_ai_waypoint.y_l = my_cv_observation.y;
	local_ai_waypoint.circle_of_acceptance = 50;
	
	/**
	 * The position of the box, the box will be picked up.
	 */	
	local_controll_struct._ai_waypoint_rel_robot.push_back(local_ai_waypoint);
      
      if( teamColor == TEAM_COLOR_BLUE ) {
	cout << "BottleBoxSort: cv_observation BOX  and TEAM_COLOR_BLUE " << endl;
	
	/**
	 * Use A* her to get the waypoints to the battery bin
	 */
	
	/**
	 * Go to waypoint
	 */
	/*
	local_ai_waypoint.x_l = 0;
	local_ai_waypoint.y_l = 0;
	local_ai_waypoint.psi_l = 0;
	local_ai_waypoint.circle_of_acceptance = 50;
	local_controll_struct._ai_waypoint.push_back(local_ai_waypoint);
	*/
	
	/**
	 * Set the blue box bin waypoint
	 */
	local_controll_struct._ai_waypoint.push_back(blue_box_bin_ai_waypoint);
        /**
	 * EraseObservation
	 */
	my_cv_observation.type = NOTHING;
	inAction = ACTIVE;
	
	priority = 100;
      }
      else if( teamColor == TEAM_COLOR_RED ) {
	cout << "BottleBoxSort: cv_observation BOX  and TEAM_COLOR_RED " << endl;
	
	/**
	 * Use A* her to get the waypoints to the box bin
	 */
	
	/**
	 * Go to waypoint
	 */
	/*
	local_ai_waypoint.x_l = 0;
	local_ai_waypoint.y_l = 0;
	local_ai_waypoint.psi_l = 0;
	local_ai_waypoint.circle_of_acceptance = 50;
	local_controll_struct._ai_waypoint.push_back(local_ai_waypoint);
	*/
	
	/**
	 * Set the red box bin waypoint
	 */
	local_controll_struct._ai_waypoint.push_back(red_box_bin_ai_waypoint);
        /**
	 * EraseObservation
	 */
	my_cv_observation.type = NOTHING;
	inAction = ACTIVE;
	
	priority = 100;
      }
      else {
	cout << "BottleBoxSort: ERROR: no teamColor BOX " << endl; 
      }
    }
    else if ( my_cv_observation.type == BOTTLE ) {
	/**
       * Set waypoints to get the BOTTLE
       * Set full priority.
       * Use the A* star to go to the bin
       * The bin position is in the box_bin_ai_waypiont struct
	 */
	
      cv_x = (float)(my_cv_observation.x);
      cv_y = (float)(my_cv_observation.y);

      local_ai_waypoint.psi_l = atan2f(cv_y, cv_x);
      local_ai_waypoint.x_l = my_cv_observation.x + 50;
      local_ai_waypoint.y_l = my_cv_observation.y;
      local_ai_waypoint.circle_of_acceptance = 50;
	
	/**
       * The position of the bottle, the bottle will be picked up.
	 */	
      local_controll_struct._ai_waypoint_rel_robot.push_back(local_ai_waypoint);
      
      if( teamColor == TEAM_COLOR_BLUE ) {
	cout << "BottleBoxSort: cv_observation BOTTLE  and TEAM_COLOR_BLUE " << endl;
	
	/**
	 * Use A* her to get the waypoints to the bottle bin
	 */
	
	/**
	 * Go to waypoint
	 */
	/*
	local_ai_waypoint.x_l = 0;
	local_ai_waypoint.y_l = 0;
	local_ai_waypoint.psi_l = 0;
	local_ai_waypoint.circle_of_acceptance = 50;
	local_controll_struct._ai_waypoint.push_back(local_ai_waypoint);
	*/
	
	/**
	 * Set the blue bottle bin waypoint
	 */
	local_controll_struct._ai_waypoint.push_back(blue_bottle_bin_ai_waypoint);
        /**
	 * EraseObservation
	 */
	my_cv_observation.type = NOTHING;
	inAction = ACTIVE;
	
	priority = 100;
      }
      else if( teamColor == TEAM_COLOR_RED ) {
	cout << "BottleBoxSort: cv_observation BOTTLE  and TEAM_COLOR_RED " << endl;
	
	/**
	 * Use A* her to get the waypoints to the bottle bin
	 */
	
	/**
	 * Go to waypoint
	 */
	/*
	local_ai_waypoint.x_l = 0;
	local_ai_waypoint.y_l = 0;
	local_ai_waypoint.psi_l = 0;
	local_ai_waypoint.circle_of_acceptance = 50;
	local_controll_struct._ai_waypoint.push_back(local_ai_waypoint);
	*/
	
	/**
	 * Set the red bottle bin waypoint
	 */
	local_controll_struct._ai_waypoint.push_back(red_bottle_bin_ai_waypoint);
        /**
	 * EraseObservation
	 */
	my_cv_observation.type = NOTHING;
	inAction = ACTIVE;
	
	priority = 100;
      }
      else {
	cout << "BottleBoxSort: ERROR: no teamColor BOTTLE " << endl; 
      }
    }
      /**
     * Proceed if not activeStrategy and no battery
     * Return -100
       */
    else if ( my_cv_observation.type == NOTHING ) {
      cout << "BottleBoxSort: my_cv_observation.type == NOTHING " << endl;
      cout << "BottleBoxSort: inAction = " << inAction << endl;
      //inAction = INACTIVE;
      return -200;
	
    }
    else {
      cout << "BottleBoxSort: my_cv_observation.type = " << my_cv_observation.type<< endl;
      cout << "BottleBoxSort: my_cv_observation.type out of range " << endl;
      cout << "BottleBoxSort: inAction = " << inAction << endl;
      //inAction = INACTIVE;
      return -200;
	
    }
  } //if( inAction == INACTIVE ) {


  cout << "BottleBoxSort: activeStrategy ( true / false ) " << activeStrategy << endl; 
  if( (inAction == ACTIVE) && (bottleBoxRelease == INACTIVE)) {
#ifdef DEBUG_BOTTLE_BOX_SORT
    cout << "\n\n\n(inAction == ACTIVE) && (bottleBoxRelease == INACTIVE)) \n\n" << endl;
#endif
    /**
     * Check if the robot is in bottle or box bin position
     */
    cout << "lastWaypointReached " << lastWaypointReached << endl;
    if( lastWaypointReached == TRUE ){
      cout << "lastWaypointReached == TRUE" << lastWaypointReached << endl;

      /**
       * Set controlstruct command to release bottle or box when in position
       */
      local_controll_struct.command.push_back( MINOR_CMD_RELEASE_BOTTLE_BOX );
      lastWaypointReached = FALSE;
      /**
       * Stop the robot and wait for the robot to release the bottle or box
       */
      local_controll_struct.setMaxSpeed = true;
      local_controll_struct.maxSpeed_mps = 0.0;
      /**
       * Very high priority
       */

      priority = 1000;

     // inAction = INACTIVE;
#ifdef DEBUG_BOTTLE_BOX_SORT
      cout << "\n\n\n\n\n" << endl;
      cout << "bottleBoxRelease = ACTIVE;" << bottleBoxRelease << endl;
      cout << "bottleBoxRelease = ACTIVE;" << bottleBoxRelease << endl;
      cout << "bottleBoxRelease = ACTIVE;" << bottleBoxRelease << endl;
      cout << "bottleBoxRelease = ACTIVE;" << bottleBoxRelease << endl;
      cout << "\n\n\n\n\n" << endl;
#endif
      bottleBoxRelease = ACTIVE;

    }
  }
  else if( (inAction == ACTIVE) && (bottleBoxRelease == ACTIVE)) {

    if( bottleBoxReleased == TRUE ){
#ifdef DEBUG_BOTTLE_BOX_SORT
      cout << "if( bottleBoxReleased == TRUE ) \n\n\n\n\n"<< endl;
      cout << "if( bottleBoxReleased == TRUE )\n\n\n\n\n\n" << endl;
#endif

      local_controll_struct.clearWaypoints = true;
      local_controll_struct.clearTheActiveStrategy = true;
      local_controll_struct.setMaxSpeed = true;
      local_controll_struct.maxSpeed_mps = MAX_SPEED_MPS;
      /**
       * Turn against bottles and boxes
       */
      if( teamColor == TEAM_COLOR_BLUE ) {
	local_ai_waypoint.x_l = blue_bottle_bin_ai_waypoint.x_l;
	local_ai_waypoint.y_l = blue_bottle_bin_ai_waypoint.y_l;
	local_ai_waypoint.psi_l = -((3*M_PI)/4);
	local_ai_waypoint.circle_of_acceptance = 100;
	local_controll_struct._ai_waypoint.push_back(local_ai_waypoint);
      }
      else if( teamColor == TEAM_COLOR_BLUE) {
	local_ai_waypoint.x_l = red_bottle_bin_ai_waypoint.x_l;
	local_ai_waypoint.y_l = red_bottle_bin_ai_waypoint.y_l;
	local_ai_waypoint.psi_l = ((1*M_PI)/4);
	local_ai_waypoint.circle_of_acceptance = 100;
	local_controll_struct._ai_waypoint.push_back(local_ai_waypoint);
	
      }
      else{
	cout << "BottleBoxSort: ERROR, no teamColor " << endl;
      }

      bottleBoxReleased = FALSE;
      bottleBoxRelease = INACTIVE;
      waitForBottleBoxReleased = 0;
    }
    else {
      waitForBottleBoxReleased = waitForBottleBoxReleased + 1;
      sleep(1);

#ifdef DEBUG_BOTTLE_BOX_SORT
      cout << " " << endl;
      cout << " " << endl;
      cout << "before waitForBottleBoxReleased = " << waitForBottleBoxReleased << endl;
      cout << " " << endl;
      cout << " " << endl;
#endif
    }
    if((waitForBottleBoxReleased < 25)) {
      waitForBottleBoxReleased = waitForBottleBoxReleased + 1;
#ifdef DEBUG_BOTTLE_BOX_SORT
      cout << " " << endl;
      cout << " " << endl;
      cout << "waitForBottleBoxReleased = " << waitForBottleBoxReleased << endl;
      cout << " " << endl;
      cout << " " << endl;
#endif
    }
    else {

      cout << " " << endl;
      cout << " " << endl;
      cout << " " << endl;
      cout << " " << endl;
      cout << " " << endl;
      cout << "BottleBoxSort: ERROR Handling " << endl;
      cout << "BottleBoxSort: timeout set BottleBoxSort inactive " << endl;
      cout << " " << endl;
      cout << " " << endl;
      cout << " " << endl;
      cout << " " << endl;
      cout << " " << endl;
      cout << " " << endl;
      agentPriority = -1;
      waitForBottleBoxReleased = 1000;

    }
  }
#ifdef DEBUG_BOTTLE_BOX_SORT
  cout << "inAction " << inAction << endl;
  cout << "batteryRelease " << bottleBoxRelease << endl;
  cout << "batteryReleased " << bottleBoxReleased << endl;
  cout << " lastWaypointReached " << lastWaypointReached << endl;
#endif

  return priority;
}



int BottleBoxSort::GetAction( controll_struct *ai_controll_struct ){
  cout << " ACTIVE STRATEGY BOTTLE_BOX_SORT " << endl;
  memcpy((void *)(ai_controll_struct),( void * )&local_controll_struct, sizeof(controll_struct));
  /**
   * Clear contol_struct
   */
  clear_controll_struct( &local_controll_struct );

  
  return 0;
}

void BottleBoxSort::Update( Subject *theChangedSubject, int typeOfObservation){
 if( theChangedSubject == _subject ){
   if( typeOfObservation == MINOR_PROXY_MESSAGE ){
     my_proxy_message = _subject->GetProxyMessage();
     if( my_proxy_message.teamColor > 0 )
      teamColor = my_proxy_message.teamColor;

   }
   else if( typeOfObservation == MINOR_CMD_PICK_UP_BOTTLE_SUCCESS ){
     typeOfPickUp = BOTTLE;
     pickUpMessageRecieved = true;
   }
   else if( typeOfObservation == MINOR_CMD_PICK_UP_BOX_SUCCESS ){
     typeOfPickUp = BOX;
     pickUpMessageRecieved = true;
   }
   else if( typeOfObservation == MINOR_CMD_PICK_UP_FAILURE ){
     typeOfPickUp = NOTHING;
     pickUpMessageRecieved = true;
   }
   else if( typeOfObservation == MINOR_NAVSYS_POSITION ){
   //  pos_obs = _subject->GetPosObservation();
   //  robot_state_3d = _subject->GetPosObservation();
     ai_waypoint_3d = _subject->GetPosObservation();
//#ifdef DEBUG_OBS_PRINT
     cout << "Type of observation POSITION BottleBoxSort" << endl;
     cout << "x_l = " << ai_waypoint_3d.x_l << endl;
     cout << "y_l = " << ai_waypoint_3d.y_l << endl;
     cout << "psi_l = " << ai_waypoint_3d.psi_l << endl;
//#endif
   }
   else if ( typeOfObservation == MINOR_NAVSYS_REACHED_WAYPOINT ) {
#ifdef DEBUG_BOTTLE_BOX_SORT
     cout << "\n\n BOTTLE BOX MINOR_NAVSYS_REACHED_WAYPOINT \n\n" << endl;
     cout << "\n\n BOTTLE BOX MINOR_NAVSYS_REACHED_WAYPOINT \n\n" << endl;
     cout << "\n\n BOTTLE BOX MINOR_NAVSYS_REACHED_WAYPOINT \n\n" << endl;
     cout << "\n\n BOTTLE BOX MINOR_NAVSYS_REACHED_WAYPOINT \n\n" << endl;
#endif
     ai_waypoint last_ai_waypoint_3d_reached;
     last_ai_waypoint_3d_reached = _subject->GetWaypointReachedObservation();
     cout << "last_ai_waypoint_3d_reached.x_l = " << last_ai_waypoint_3d_reached.x_l << endl;
     cout << "last_ai_waypoint_3d_reached.y_l = " << last_ai_waypoint_3d_reached.y_l << endl;
     if(teamColor == TEAM_COLOR_BLUE) {
       if((last_ai_waypoint_3d_reached.x_l == blue_bottle_bin_ai_waypoint.x_l) && (last_ai_waypoint_3d_reached.y_l ==  blue_bottle_bin_ai_waypoint.y_l))
	 lastWaypointReached = TRUE;
       else if ((last_ai_waypoint_3d_reached.x_l == blue_box_bin_ai_waypoint.x_l) && (last_ai_waypoint_3d_reached.y_l == blue_box_bin_ai_waypoint.y_l))
	 lastWaypointReached = TRUE;
     }
     else if (teamColor == TEAM_COLOR_RED) {
       if((last_ai_waypoint_3d_reached.x_l == red_bottle_bin_ai_waypoint.x_l) && (last_ai_waypoint_3d_reached.y_l == red_bottle_bin_ai_waypoint.y_l))
	 lastWaypointReached = TRUE;
       else if ((last_ai_waypoint_3d_reached.x_l == red_box_bin_ai_waypoint.x_l) && (last_ai_waypoint_3d_reached.y_l == red_box_bin_ai_waypoint.y_l))
	 lastWaypointReached = TRUE;
     }
#ifdef DEBUG_BOTTLE_BOX_SORT
     if(lastWaypointReached == TRUE)
      cout << "\n \n \n \n BottleBoxSort: BOTTLE_BOX_BIN REACHED \n \n \n \n \n " << endl;
#endif
   }
   else if ( typeOfObservation == MINOR_CMD_BOTTLE_BOX_RELEASED ) {
     bottleBoxReleased = TRUE;
#ifdef DEBUG_BOTTLE_BOX_SORT
     cout << "\n \n \n \n BottleBoxSort: bottleboxReleased = TRUE; \n \n \n \n \n " << endl;
     cout << "\n \n \n \n BottleBoxSort: bottleBoxReleased = TRUE; \n \n \n \n \n " << endl;
     cout << "\n \n \n \n BottleBoxSort: bottleBoxReleased = TRUE; \n \n \n \n \n " << endl;
#endif
   }

   else if( typeOfObservation == MINOR_CV_OBSERVATION ){
     my_cv_observation = _subject->GetCvObservation();
#ifdef DEBUG_OBS_PRINT
     cout << "Type of observation CV BottleBoxSort" << endl;
     cout << "my_cv_observation.type = " << my_cv_observation.type << endl; 
     cout << "my_cv_observation.x = " << my_cv_observation.x << endl; 
     cout << "my_cv_observation.y = " << my_cv_observation.y << endl;
     cout << "Type of observation CV BottleBoxSort" << endl;
     cout << "my_cv_observation.type = " << my_cv_observation.type << endl; 
     cout << "my_cv_observation.x = " << my_cv_observation.x << endl; 
     cout << "my_cv_observation.y = " << my_cv_observation.y << endl;
     cout << "Type of observation CV BottleBoxSort" << endl;
     cout << "my_cv_observation.type = " << my_cv_observation.type << endl; 
     cout << "my_cv_observation.x = " << my_cv_observation.x << endl; 
     cout << "my_cv_observation.y = " << my_cv_observation.y << endl;

#endif
   }
#ifdef DEBUG_OBS_PRINT
 //  cout << "Update BottleBoxSort" << endl;
 //  cout << " " << endl;
#endif


 }
}

/**
 * The strategy is interrupted by anoter strategy
 * keep existing settings
 */
int BottleBoxSort::StrategyInterrupted() {

  
  return 0;
}

int BottleBoxSort::GetInterruptAction(interrupt_controll_struct *interrupt_controll) {
  
  return 0; 
}


