/**
 *  Eurobot 2007
 *  PlansysObservation.cpp
 *  Gunnar Kjemphol <gunnar@kjemphol.no>
 */

#include "BottleBoxSort.h"


BottleBoxSort::BottleBoxSort ( PlansysObservation *s , int theTeamColor ){
  teamColor = theTeamColor;
  /**
   * Give the Agent a priority in the constructor [0-9]
   * if the agentPriority is -1 the agent is inactive
   */
  agentPriority = 5;
  inAction = INACTIVE;
  activeStrategy = INACTIVE;
  bottleBoxRelease = INACTIVE;
  bottleBoxReleased = FALSE;
  lastWaypointReached = FALSE;
  pickUpWaypointReached = FALSE;
  pickUpMessageSendt = FALSE;
  lastWaypointSendt  = FALSE;
  changeTheObservationType = FALSE;
  turnAgainstBottlesAndBoxes = FALSE;
  waitForBottleBoxReleased = 0;
  typeOfPickUp = NOTHING;
  pickUpMessageRecieved = false;
  lastWaypointId = 0;
  pickUpWaypointId = 0;
  _subject = s;
  _subject->Attach(this);

  /**
   * Bottle and bin positions
   * Two for each teamColor
   */

    red_bottle_bin_ai_waypoint.x_l = 700;
    red_bottle_bin_ai_waypoint.y_l = 1600;
    red_bottle_bin_ai_waypoint.psi_l = ((7*M_PI)/8);
    red_bottle_bin_ai_waypoint.circle_of_acceptance = 150;

    red_box_bin_ai_waypoint.x_l = 500;
    red_box_bin_ai_waypoint.y_l = 1600;
    red_box_bin_ai_waypoint.psi_l = ((5*M_PI)/8);
    red_box_bin_ai_waypoint.circle_of_acceptance = 150;


    blue_bottle_bin_ai_waypoint.x_l = 2300; //1500
    blue_bottle_bin_ai_waypoint.y_l = 1600; //1050
    blue_bottle_bin_ai_waypoint.psi_l = ((1*M_PI)/8);
    blue_bottle_bin_ai_waypoint.circle_of_acceptance = 150;

    blue_box_bin_ai_waypoint.x_l = 2600; //1500
    blue_box_bin_ai_waypoint.y_l = 1600; //1050
    blue_box_bin_ai_waypoint.psi_l = ((7*M_PI)/16);
    blue_box_bin_ai_waypoint.circle_of_acceptance = 150;

    /**
     * Error team color, ask for team color
     */

  clear_controll_struct(&local_controll_struct);
  my_cv_observation.x = 0;
  my_cv_observation.y = 0;
  my_cv_observation.type = NOTHING;
  
  color_cv_observation.x = 0;
  color_cv_observation.y = 0;
  color_cv_observation.type = NOTHING;

}



BottleBoxSort::~BottleBoxSort (){
  _subject->Detach(this);

}

int BottleBoxSort::AgentGlobalPriority(){
  return agentPriority; 
}

int BottleBoxSort::SetAgentActive( int active_r ){
  inAction = active_r;
  activeStrategy = ACTIVE;
  if( active_r == INACTIVE) {
    agentPriority = 5;
    inAction = INACTIVE;
    activeStrategy = INACTIVE;
    bottleBoxRelease = INACTIVE;
    bottleBoxReleased = FALSE;
    lastWaypointReached = FALSE;
    pickUpWaypointReached = FALSE;
    pickUpMessageSendt = FALSE;
    lastWaypointSendt  = FALSE;
    turnAgainstBottlesAndBoxes = FALSE;
    waitForBottleBoxReleased = 0;
    typeOfPickUp = NOTHING;
    pickUpMessageRecieved = false;
    color_cv_observation.type = NOTHING;
    color_cv_observation.x = 0;
    color_cv_observation.y = 0;
    my_cv_observation.type = NOTHING;
    my_cv_observation.x = 0;
    my_cv_observation.y = 0;
    clear_controll_struct(&local_controll_struct);
    
  }
}



int BottleBoxSort::GetPriority(){
  int priority = -100;
  int i;
  float cv_x, cv_y;
  /**
   * Lock for battery
   */
#ifdef DEBUG_BOTTLE_BOX_SORT
  cout << " " << endl;
  cout << "BOTTLE_BOX_SORT: Strategy Active = " << activeStrategy << endl;
  cout << " " << endl;
  cout << "BOTTLE_BOX_SORT: inAction " << inAction << endl;
  cout << "BOTTLE_BOX_SORT: pickUpWaypointReached " << pickUpWaypointReached << endl;
  cout << "BOTTLE_BOX_SORT: lastWaypointSendt " << lastWaypointSendt << endl;
  cout << "BOTTLE_BOX_SORT: lastWaypointReached " << lastWaypointReached << endl;
  cout << "BOTTLE_BOX_SORT: my_cv_observation.type " << my_cv_observation.type << endl;
  cout << "BOTTLE_BOX_SORT: my_cv_observation.y " << my_cv_observation.x << endl;
  cout << "BOTTLE_BOX_SORT: my_cv_observation.y " << my_cv_observation.y << endl;
  cout << "current_bin_ai_waypoint.x_l " << current_bin_ai_waypoint.x_l << endl;
  cout << "current_bin_ai_waypoint.y_l " << current_bin_ai_waypoint.y_l << endl;
  cout << " \n\n " << endl;
#endif
  if ( inAction == ACTIVE ) {
   priority = 100;
  }
  if( inAction == INACTIVE ) {
    cout << "BOTTLE_BOX_SORT: inAction == INACTIVE << " << endl;

    if( my_cv_observation.type == BOX ){
      /**
       * Set waypoints to get the BOX
       * Set full priority.
       */
	
      cv_x = (float)(my_cv_observation.x);
      cv_y = (float)(my_cv_observation.y);

      pick_up_waypoint.psi_l = atan2f(cv_y, cv_x);//0.0;
      pick_up_waypoint.x_l = my_cv_observation.x + 100; // Add 50mm to be sure
      pick_up_waypoint.y_l = my_cv_observation.y;
      pick_up_waypoint.id = 10;
      pick_up_waypoint.circle_of_acceptance = 50;
	
      /**
       * The position of the box, the box will be picked up.
       */	
      local_controll_struct._ai_waypoint_rel_robot.push_back(pick_up_waypoint);
      
      inAction = ACTIVE;
      priority = 500;

    }
    else if( my_cv_observation.type == BOTTLE ){
      /**
       * Set waypoints to get the BOX
       * Set full priority.
       */
	
      cv_x = (float)(my_cv_observation.x);
      cv_y = (float)(my_cv_observation.y);

      pick_up_waypoint.psi_l = atan2f(cv_y, cv_x);
      pick_up_waypoint.x_l = my_cv_observation.x + 100; // Add 50mm to be sure
      pick_up_waypoint.y_l = my_cv_observation.y;
      pick_up_waypoint.id = 10;
      pick_up_waypoint.circle_of_acceptance = 50;
	
      /**
       * The position of the bottle, the bottle will be picked up.
       */	
      local_controll_struct._ai_waypoint_rel_robot.push_back(pick_up_waypoint);
      
      inAction = ACTIVE;
      priority = 500;

    }
    else if( my_cv_observation.type == NOTHING ){
#ifndef DEBUG_BOTTLE_BOX_SORT
      cout << "BottleBoxSort my_cv_observation.type NOTHING " << endl;
#endif
      priority = -100;
      return priority;
      
    }
    else {
      cout << "BottleBoxSort my_cv_observation.type wrong type " << endl;
      priority = -100;
      return priority;
    }
  }
  /**
   * The picked up item is another type then first expected 
   * change the bin waypoint, and send the new waypoint to navsys
   */
  if (changeTheObservationType == TRUE) {
    if( color_cv_observation.type == BOX ) {
      if(teamColor == TEAM_COLOR_BLUE ) {
	memcpy((void *)(&current_bin_ai_waypoint), (void *)(&blue_box_bin_ai_waypoint), sizeof(ai_waypoint));
	local_controll_struct._ai_waypoint.push_back(blue_box_bin_ai_waypoint);
      }
      else if(teamColor == TEAM_COLOR_RED) {
	memcpy((void *)(&current_bin_ai_waypoint), (void *)(&red_box_bin_ai_waypoint), sizeof(ai_waypoint));
	local_controll_struct._ai_waypoint.push_back(red_box_bin_ai_waypoint);
      }
      lastWaypointSendt = TRUE;
    }
    else if( color_cv_observation.type == BOTTLE ) {
      if(teamColor == TEAM_COLOR_BLUE ) {
	memcpy((void *)(&current_bin_ai_waypoint), (void *)(&blue_bottle_bin_ai_waypoint), sizeof(ai_waypoint));
	local_controll_struct._ai_waypoint.push_back(blue_bottle_bin_ai_waypoint);
      }
      else if(teamColor == TEAM_COLOR_RED) {
	memcpy((void *)(&current_bin_ai_waypoint), (void *)(&red_bottle_bin_ai_waypoint), sizeof(ai_waypoint));
	local_controll_struct._ai_waypoint.push_back(red_bottle_bin_ai_waypoint);
      }
      lastWaypointSendt = TRUE;
    }

    changeTheObservationType = FALSE;
    priority = 500;
    return priority;
  }

  /**
   * Check if pick_up_waypoint is reached
   */
  if( (inAction == ACTIVE) && ( pickUpWaypointReached == TRUE) && (pickUpMessageSendt == FALSE)) {
    /**
     * Send pick up mesage
     */
    cout << "pick up " << endl;
    local_controll_struct.command.push_back( MINOR_CMD_PICK_UP_BOTTLE_BOX );
    pickUpMessageSendt = TRUE;
    /**
     * Send message to CVsys, asc for botte og box
     */
    priority = 600;
    return priority;

  }

  /**
   * Go to waypoint bin. Maybe wait for the pickup unit to confirm pickup, 
   * or maybe not necessarely
   */
  if( (inAction == ACTIVE) && ( pickUpWaypointReached == TRUE) && (pickUpMessageSendt == TRUE) && (lastWaypointSendt == FALSE) && (lastWaypointReached == FALSE) ) {
    if( my_cv_observation.type == BOX ) {
      if(teamColor == TEAM_COLOR_BLUE ) {
	memcpy((void *)(&current_bin_ai_waypoint), (void *)(&blue_box_bin_ai_waypoint), sizeof(ai_waypoint));
	local_controll_struct._ai_waypoint.push_back(blue_box_bin_ai_waypoint);
      }
      else if(teamColor == TEAM_COLOR_RED) {
	memcpy((void *)(&current_bin_ai_waypoint), (void *)(&red_box_bin_ai_waypoint), sizeof(ai_waypoint));
	local_controll_struct._ai_waypoint.push_back(red_box_bin_ai_waypoint);
      }
      lastWaypointSendt = TRUE;
    }
    else if( my_cv_observation.type == BOTTLE ) {
      if(teamColor == TEAM_COLOR_BLUE ) {
	memcpy((void *)(&current_bin_ai_waypoint), (void *)(&blue_bottle_bin_ai_waypoint), sizeof(ai_waypoint));
	local_controll_struct._ai_waypoint.push_back(blue_bottle_bin_ai_waypoint);
      }
      else if(teamColor == TEAM_COLOR_RED) {
	memcpy((void *)(&current_bin_ai_waypoint), (void *)(&red_bottle_bin_ai_waypoint), sizeof(ai_waypoint));
	local_controll_struct._ai_waypoint.push_back(red_bottle_bin_ai_waypoint);
      }
      lastWaypointSendt = TRUE;
    }
    
    priority = 700;
    return priority;
  }
  
  /**
   * Wait for bottle/box bin reached, and release bottle box
   */
  if( (lastWaypointReached == TRUE) && (bottleBoxReleased == FALSE) ) {
    cout << "Release Bottle or Box " << endl;
    local_controll_struct.command.push_back( MINOR_CMD_RELEASE_BOTTLE_BOX );
    bottleBoxReleased = TRUE;
    priority = 800;
    return priority;
  }
  
  if( bottleBoxReleased == TRUE && turnAgainstBottlesAndBoxes == FALSE ){
    turnAgainstBottlesAndBoxes = TRUE;
   /**
    * Turn against bottles and boxes
    */
    if( teamColor == TEAM_COLOR_BLUE ) {
      local_ai_waypoint.x_l = blue_bottle_bin_ai_waypoint.x_l - 150;
      local_ai_waypoint.y_l = blue_bottle_bin_ai_waypoint.y_l - 150;
      local_ai_waypoint.psi_l = -((3*M_PI)/4);
      local_ai_waypoint.circle_of_acceptance = 100;
      local_controll_struct._ai_waypoint.push_back(local_ai_waypoint);
    }
    else if( teamColor == TEAM_COLOR_RED) {
      local_ai_waypoint.x_l = red_bottle_bin_ai_waypoint.x_l + 150;
      local_ai_waypoint.y_l = red_bottle_bin_ai_waypoint.y_l - 150;
      local_ai_waypoint.psi_l = -((1*M_PI)/4);
      local_ai_waypoint.circle_of_acceptance = 100;
      local_controll_struct._ai_waypoint.push_back(local_ai_waypoint);
    }
  }
  /**
   *  Bottle or Box is released clear the strategy
   */
  else if( bottleBoxReleased == TRUE && turnAgainstBottlesAndBoxes == TRUE ) {

    SetAgentActive( INACTIVE );
    local_controll_struct.clearTheActiveStrategy = TRUE;
    priority = 1000;
    return priority;

  }

  return priority;
}


int BottleBoxSort::GetAction( controll_struct *ai_controll_struct ){
  if(inAction == INACTIVE)
    cout << "BottleBoxSort::GetAction, inAction == INACTIVE  " << endl;
  cout << " ACTIVE STRATEGY BOTTLE_BOX_SORT " << endl;
  memcpy((void *)(ai_controll_struct),( void * )&local_controll_struct, sizeof(controll_struct));
  /**
   * Clear contol_struct
   */
  clear_controll_struct( &local_controll_struct );

  
  return 0;
}

void BottleBoxSort::Update( Subject *theChangedSubject, int typeOfObservation){
 if( theChangedSubject == _subject ){
   if( typeOfObservation == MINOR_PROXY_MESSAGE ){
     my_proxy_message = _subject->GetProxyMessage();
     if( my_proxy_message.teamColor > 0 )
      teamColor = my_proxy_message.teamColor;

   }
   else if( typeOfObservation == MINOR_CMD_PICK_UP_BOTTLE_SUCCESS ){
     typeOfPickUp = BOTTLE;
     pickUpMessageRecieved = true;
   }
   else if( typeOfObservation == MINOR_CMD_PICK_UP_BOX_SUCCESS ){
     typeOfPickUp = BOX;
     pickUpMessageRecieved = true;
   }
   else if( typeOfObservation == MINOR_CMD_PICK_UP_FAILURE ){
     typeOfPickUp = NOTHING;
     pickUpMessageRecieved = true;
   }
   else if( typeOfObservation == MINOR_NAVSYS_POSITION ){

     ai_waypoint_3d = _subject->GetPosObservation();
#ifdef DEBUG_OBS_PRINT
     cout << "Type of observation POSITION BottleBoxSort" << endl;
     cout << "x_l = " << ai_waypoint_3d.x_l << endl;
     cout << "y_l = " << ai_waypoint_3d.y_l << endl;
     cout << "psi_l = " << ai_waypoint_3d.psi_l << endl;
#endif
   }
   else if ( typeOfObservation == MINOR_NAVSYS_REACHED_WAYPOINT ) {
#ifdef DEBUG_BOTTLE_BOX_SORT
     cout << "\n\n BOTTLE BOX MINOR_NAVSYS_REACHED_WAYPOINT \n\n" << endl;
     cout << "\n\n BOTTLE BOX MINOR_NAVSYS_REACHED_WAYPOINT \n\n" << endl;
     cout << "\n\n BOTTLE BOX MINOR_NAVSYS_REACHED_WAYPOINT \n\n" << endl;
     cout << "\n\n BOTTLE BOX MINOR_NAVSYS_REACHED_WAYPOINT \n\n" << endl;
#endif
     ai_waypoint last_ai_waypoint_reached;
     last_ai_waypoint_reached = _subject->GetWaypointReachedObservation();
     cout << "last_ai_waypoint_reached.x_l = " << last_ai_waypoint_reached.x_l << endl;
     cout << "last_ai_waypoint_reached.y_l = " << last_ai_waypoint_reached.y_l << endl;
     cout << "last_ai_waypoint_reached.id = " << last_ai_waypoint_reached.id << endl;
     cout << "pick_up_waypoint.x_l = " << pick_up_waypoint.x_l << endl;
     cout << "pick_up_waypoint.y_l = " << pick_up_waypoint.y_l << endl;
     cout << "pick_up_waypoint.id = " << pick_up_waypoint.id << endl;
  //   while(1);
     /**
      * Waypoint pick up bottle box position
      * pick_up_waypoint
      */
     if( pick_up_waypoint.id == last_ai_waypoint_reached.id) {

       pickUpWaypointReached = TRUE;

     }
     
     if((current_bin_ai_waypoint.x_l >= last_ai_waypoint_reached.x_l - 10) &&
	 (current_bin_ai_waypoint.x_l <= last_ai_waypoint_reached.x_l + 10) &&
	 (current_bin_ai_waypoint.y_l >= last_ai_waypoint_reached.y_l - 10) && 
	 (current_bin_ai_waypoint.y_l <= last_ai_waypoint_reached.y_l + 10) ){
       
       lastWaypointReached = TRUE;
     }
       

     /*
     if(teamColor == TEAM_COLOR_BLUE) {
       if((last_ai_waypoint_reached.x_l == blue_bottle_bin_ai_waypoint.x_l) && (last_ai_waypoint_reached.y_l ==  blue_bottle_bin_ai_waypoint.y_l))
	 lastWaypointReached = TRUE;
       else if ((last_ai_waypoint_reached.x_l == blue_box_bin_ai_waypoint.x_l) && (last_ai_waypoint_reached.y_l == blue_box_bin_ai_waypoint.y_l))
	 lastWaypointReached = TRUE;
     }
     else if (teamColor == TEAM_COLOR_RED) {
       if((last_ai_waypoint_reached.x_l == red_bottle_bin_ai_waypoint.x_l) && (last_ai_waypoint_reached.y_l == red_bottle_bin_ai_waypoint.y_l))
	 lastWaypointReached = TRUE;
       else if ((last_ai_waypoint_reached.x_l == red_box_bin_ai_waypoint.x_l) && (last_ai_waypoint_reached.y_l == red_box_bin_ai_waypoint.y_l))
	 lastWaypointReached = TRUE;
     }
     */
#ifdef DEBUG_BOTTLE_BOX_SORT
     if(lastWaypointReached == TRUE)
      cout << "\n \n \n \n BottleBoxSort: BOTTLE_BOX_BIN OR PICK UP REACHED \n \n \n \n \n " << endl;
#endif
   }
   else if ( typeOfObservation == MINOR_CMD_BOTTLE_BOX_RELEASED ) {
     bottleBoxReleased = TRUE;
#ifdef DEBUG_BOTTLE_BOX_SORT
     cout << "\n \n \n \n BottleBoxSort: bottleboxReleased = TRUE; \n \n \n \n \n " << endl;
#endif
   }

   else if( typeOfObservation == MINOR_CV_OBSERVATION ){
     cv_observation temp_cv_observation;
     temp_cv_observation = _subject->GetCvObservation();
     if(temp_cv_observation.x == 0){
     }
     else
       memcpy((void *)(&my_cv_observation),( void * )&temp_cv_observation, sizeof(cv_observation));
       
#ifdef DEBUG_OBS_PRINT
     cout << "Type of observation CV BottleBoxSort" << endl;
     cout << "my_cv_observation.type = " << my_cv_observation.type << endl; 
     cout << "my_cv_observation.x = " << my_cv_observation.x << endl; 
     cout << "my_cv_observation.y = " << my_cv_observation.y << endl;
#endif
   }
   else if( typeOfObservation == MINOR_CV_COLOR_OBSERVATION ){
     color_cv_observation = _subject->GetCvColorObservation();
     if(my_cv_observation.type == color_cv_observation.type ){
       /**
        * Do nothing, the pick up is correct
        */
    }
    else {
     /**
      * Change the pick up type, and change last waypoint. In the getPriority 
      * function send the new waypoint
      */
      my_cv_observation.type = color_cv_observation.type;
      changeTheObservationType = TRUE;
#ifdef DEBUG_OBS_PRINT
      cout << "Type of observation CV_COLOR BottleBoxSort" << endl;
      cout << "my_cv_observation.type = " << my_cv_observation.type << endl; 
      cout << "my_cv_observation.x = " << my_cv_observation.x << endl; 
      cout << "my_cv_observation.y = " << my_cv_observation.y << endl;
#endif
    }
   }
#ifdef DEBUG_OBS_PRINT
 //  cout << "Update BottleBoxSort" << endl;
 //  cout << " " << endl;
#endif


 }
}

/**
 * The strategy is interrupted by anoter strategy
 * keep existing settings
 */
int BottleBoxSort::StrategyInterrupted() {

  
  return 0;
}

int BottleBoxSort::GetInterruptAction(interrupt_controll_struct *interrupt_controll) {
  
  return 0; 
}


