#ifndef EMPTY_STRATEGY_H
#define EMPTY_STRATEGY_H

/**
 *  Eurobot 2007
 *  EmptyStrategy.h
 *  author "Gunnar Kjemphol <gunnar@kjemphol.no>"
 * 
 * 
 *  This is a child class of Observer class
 * 
 *  The PlansysObservation objects can call EmptyStrategy memberfunction 
 *  Update() when a new observation is recieved. 
 * 
 * 
 */

#include <math.h>

#include "Observer.h"
#include "PlansysObservation.h"
#include "mainplansys.h"
#include "waypoint.h"
#include "posix_message_config.h"
#include "observations.h"
#include "init.h"


class EmptyStrategy: public Observer {
  public:
    EmptyStrategy( PlansysObservation * , int theTeamColor );
    ~EmptyStrategy();
    void Update( Subject *, int typeOfObservation);
    int GetPriority();
    int AgentGlobalPriority();
    int GetAction( controll_struct *ai_controll_struct );
    int SetAgentActive( int active );
    int StrategyInterrupted();
    int GetInterruptAction(interrupt_controll_struct *interrupt_controll);

  private:
  /**
   * Give the Agent a priority in the constructor [0-9]
   * if the agentPriority is -1 the agent is inactive
   */
    int agentPriority;
    int inAction;
    int activeStrategy;

    PlansysObservation *_subject;
    ai_waypoint ai_waypoint_3d;             // Position
    ai_waypoint ai_waypoint_reached;        // Last waypoint reached
    cv_observation my_cv_observation;       // Last CV observation
    cv_observation color_cv_observation;    // CV color observation
    proxy_message my_proxy_message;         // Proxy Message
    controll_struct local_controll_struct;  // Control Struct


};

#endif
