/**
 *  Eurobot 2007
 *  MotorAi.cpp
 *  Gunnar Kjemphol <gunnar@kjemphol.no>
 */

#include "MotorAi.h"
#include <sys/time.h>    
#include <math.h>


MotorAi::MotorAi ( PlansysObservation *s , int theTeamColor ) {
  /**
   * Give the Agent a priority in the constructor [0-9]
   * if the agentPriority is -1 the agent is inactive
   */
  teamColor = theTeamColor;
  agentPriority = 9;
  inAction = INACTIVE;
  activeStrategy = INACTIVE;
  _subject = s;
  _subject->Attach(this);

  clear_controll_struct(&local_controll_struct);
  
  gettimeofday(&collision_timer,NULL);
  
  collision_time_value_1 = collision_timer.tv_sec;
  collision_time_value_2 = collision_time_value_1;
  collision_counter = 0;

}

MotorAi::~MotorAi (){
  _subject->Detach(this);

}

int MotorAi::AgentGlobalPriority(){
  return agentPriority; 
}

int MotorAi::SetAgentActive( int active ){
  activeStrategy = active;
  if ( active == INACTIVE)
    inAction = INACTIVE;
}


int MotorAi::GetPriority(){
  
  
  
  return -10;
}

int MotorAi::GetAction( controll_struct *ai_controll_struct ){
  cout << " ACTIVE STRATEGY MOTOR_AI " << endl;
  memcpy((void *)(ai_controll_struct),( void * )&local_controll_struct, sizeof(controll_struct));
  /**
   * Clear contol_struct
   */
  clear_controll_struct( &local_controll_struct );

  return -10;
}

void MotorAi::Update( Subject *theChangedSubject, int typeOfObservation ){
 if( theChangedSubject == _subject ){
   if( typeOfObservation == MINOR_COLLISION_DETECTED ){
     collision_message = _subject->GetCollisionMessage();

   }
   
   else if( typeOfObservation == MINOR_PROXY_MESSAGE ){
     my_proxy_message = _subject->GetProxyMessage();
     if( my_proxy_message.teamColor > 0 )
       teamColor = my_proxy_message.teamColor;

   }
   else if(typeOfObservation == MINOR_NAVSYS_POSITION){
//     robot_state_3d = _subject->GetPosObservation();
     ai_waypoint_3d = _subject->GetPosObservation(); 
     //pos_obs = _subject->GetPosObservation();
#ifdef DEBUG_OBS_PRINT
     cout << " " << endl;
     cout << "Type of observation POSITION MotorAi" << endl;
     cout << "x_l = " << ai_waypoint_3d.x_l << endl;
     cout << "y_l = " << ai_waypoint_3d.y_l << endl;
     cout << "u_b = " << ai_waypoint_3d.psi_l << endl;
#endif
   }
#ifdef DEBUG_OBS_PRINT
   cout << "Update MotorAi" << endl;
   cout << " " << endl;
#endif

 }
}

/**
 * The strategy is interrupted by anoter strategy
 * keep existing settings
 */
int MotorAi::StrategyInterrupted() {


}

int MotorAi::GetInterruptAction(interrupt_controll_struct *interrupt_controll) {
  int i;
  ai_waypoint interrupt_waypoint;
//  gettimeofday(&collision_timer,NULL);
//  collision_time_value_1 = collision_timer.tv_sec;
  
//  collision_time_value_2 = collision_time_value_1;

//  if(collision_time_value_1 > collision_time_value_2 + 2){
//    collision_time_value_2 = collision_time_value_1;



    if(collision_message.data[2] < 0x20 && collision_message.data[2] > 0x01){
      interrupt_controll->stopRobot = true;
      interrupt_waypoint.x_l = 0;
      interrupt_waypoint.y_l = 350;
      interrupt_waypoint.circle_of_acceptance = 50;
      interrupt_waypoint.psi_l = M_PI;

      interrupt_controll->_interrupt_ai_waypoint.push_back(interrupt_waypoint);
    }

    else if(collision_message.data[0] < 0x20 && collision_message.data[0] > 0x01){
      interrupt_controll->stopRobot = true;
      interrupt_waypoint.x_l = 0;
      interrupt_waypoint.y_l = -350;
      interrupt_waypoint.circle_of_acceptance = 50;
      interrupt_waypoint.psi_l = M_PI;

      interrupt_controll->_interrupt_ai_waypoint.push_back(interrupt_waypoint);

    }
    else if(collision_message.data[3] < 0x20 && collision_message.data[3] > 0x01){
      interrupt_controll->stopRobot = true;

      interrupt_waypoint.x_l = -350;
      interrupt_waypoint.y_l = 350;
      interrupt_waypoint.circle_of_acceptance = 50;
      interrupt_waypoint.psi_l = -M_PI;

      interrupt_controll->_interrupt_ai_waypoint.push_back(interrupt_waypoint);

    }

//  }
  /*
  if(collision_message.data[4] < 0x20 && collision_message.data[4] > 0x01){
    interrupt_controll->stopRobot = true;
  }
  if(collision_message.data[5] < 0x20 && collision_message.data[5] > 0x01){
    interrupt_controll->stopRobot = true;
  }
  */
  
  /**
   * Clear the internal collisoin_message
   */
  for(i = 0; i < 8; i++)
    collision_message.data[i] = 0x00;
  
  return 0;
}



