#ifndef MOTOR_AI_H
#define MOTOR_AI_H

/**
 *  Eurobot 2007
 *  MotorAi.h
 *  Gunnar Kjemphol <gunnar@kjemphol.no>
 * 
 * 
 *  This is a child class of Observer class
 * 
 *  The PlansysObservation objects can call MotorAi memberfunction 
 *  Update() when a new observation is recieved. 
 * 
 * This function will check for motor stall.
 * 
 */

#include "Observer.h"
#include "PlansysObservation.h"
#include "mainplansys.h"
#include "waypoint.h"
#include "posix_message_config.h"
#include "init.h"

class MotorAi: public Observer {
  public:
    MotorAi( PlansysObservation * , int theTeamColor );
    ~MotorAi();
    void Update( Subject *, int typeOfObservation);
    int GetPriority();
    int AgentGlobalPriority();
    int GetAction( controll_struct *ai_controll_struct );
    int SetAgentActive( int active );
    int StrategyInterrupted();
    int GetInterruptAction(interrupt_controll_struct *interrupt_controll);


  private:
  /**
   * Give the Agent a priority in the constructor [0-9]
   * if the agentPriority is -1 the agent is inactive in priority
   */

    int agentPriority;
    int inAction;
    int activeStrategy;
    int lastWaypointReached;
    PlansysObservation *_subject; // referece to subject
//    robot_state_3dof_posix robot_state_3d;
    ai_waypoint ai_waypoint_3d;
    proxy_message my_proxy_message;
    cv_observation my_cv_observation;
    collision_detection_message collision_message;
    controll_struct local_controll_struct;

    timeval collision_timer;
    double collision_time_value_1;
    double collision_time_value_2;
    int collision_counter;

};

#endif
