/**
 *  Eurobot 2007
 *  PlansysObservation.h
 *  Gunnar Kjemphol <gunnar@kjemphol.no>
 * 
 */

// Includes
#include "PlansysObservation.h"

PlansysObservation::PlansysObservation(){
  activeStrategy = INACTIVE;
  activeStrategyValue = -1;
  strategyCandidate = -1;
#ifdef DEBUG_INIT_PRINT
  cout << "Plansys constructor" << endl;
#endif

  ai_waypoint_3d_ptr = &ai_waypoint_3d;
  
  last_ai_waypoint_3d_reached.x_l = -1;
  last_ai_waypoint_3d_reached.y_l = -1;
  last_ai_waypoint_3d_reached.psi_l = 0.0;
  last_ai_waypoint_3d_reached.circle_of_acceptance = 20;
  
}

int PlansysObservation::NewObservation(pmq_message pm, int typeOfObservation) {
  if( typeOfObservation == MINOR_COLLISION_DETECTED ) {
    memcpy((void *)(&collision_message),( void * )pm.data, sizeof(collision_detection_message));

  }
  else if( typeOfObservation == MINOR_NAVSYS_POSITION ) {
#ifdef DEBUG_OBS_PRINT
    cout << " " << endl;
    cout << "Recieve Position Update" << endl;
#endif
    // Overflødig opprettelse av struct
 posix_robot_state_3dof pos_robot;
 memcpy((void *)(&pos_robot),( void * )pm.data, sizeof(posix_robot_state_3dof));
 ai_waypoint_3d.x_l = pos_robot.x_l;
 ai_waypoint_3d.y_l = pos_robot.y_l;
 ai_waypoint_3d.psi_l = pos_robot.psi_l;
 ai_waypoint_3d.circle_of_acceptance = 0;

  }
  else if( typeOfObservation == MINOR_NAVSYS_REACHED_WAYPOINT ) {
#ifdef DEBUG_OBS_PRINT
    cout << " " << endl;
    cout << "PlansysObservation: MINOR_NAVSYS_REACHED_WAYPOINT" << endl;

#endif
    posix_robot_state_3dof pos_robot;
    memcpy((void *)(&last_ai_waypoint_3d_reached),( void * )pm.data, sizeof(ai_waypoint));


  }
  else if (typeOfObservation == MINOR_CV_OBSERVATION ) {
#ifdef DEBUG_OBS_PRINT
    cout << " " << endl;
    cout << "Recieve Cv observation" << endl;
#endif
    memcpy((void *)(&my_cv_observation),( void * )pm.data, sizeof(cv_observation));
#ifdef DEBUG_OBS_PRINT
    cout << " PlansysObservation: my_cv_observation.type " << my_cv_observation.type << endl;
    cout << " PlansysObservation: my_cv_observation.x " << my_cv_observation.x << endl;
    cout << " PlansysObservation: my_cv_observation.y " << my_cv_observation.y << endl;
    cout << " PlansysObservation: my_cv_observation.ret " << my_cv_observation.ret << endl;
#endif


  }
  else if (typeOfObservation == MINOR_CV_COLOR_OBSERVATION ) {
#ifdef DEBUG_OBS_PRINT
    cout << " " << endl;
    cout << "Recieve Cv COLOR observation" << endl;
#endif
    memcpy((void *)(&color_cv_observation),( void * )pm.data, sizeof(cv_observation));
#ifdef DEBUG_OBS_PRINT
    cout << " PlansysObservation: color_cv_observation.type " << color_cv_observation.type << endl;
#endif


  }
  else if (typeOfObservation == GAME_TIME_MESSAGE ) {
#ifdef DEBUG_TIME
    cout << " " << endl;
    cout << "Recieve TIME observation" << endl;
    
#endif
    main_time.timeLeft = pm.length; // Uses length for time
  }

  Notify( typeOfObservation );

}



void PlansysObservation::GetAgentPriority(){

  int i, x, y , max;
  vector<int> tempStrategyValueVector;
  int tempStategyValue = 0;
  int tempStrategyEmergencyValue = 0;
  int tempStrategy = -1;
  int tempStrategyEmergency = -1;
  max = 0;
  /**
   * Agent Number for the active strategy
   */


  for( i = 0; i < _observers.size(); i++ ) {
    x = (_observers.at(i))->GetPriority();
    y = (_observers.at(i))->AgentGlobalPriority();
#ifdef DEBUG_ACTION_PRINT
    cout << "**** ACTIVE STRATERY ***** = " << activeStrategy << endl;
    cout << "observer number " << i << endl;
    cout << "x = (_observers.at(i))->GetPriority();" << x << endl;
    cout << "y = (_observers.at(i))->AgentGlobalPriority();" << y << endl;
#endif
    tempStrategyValueVector.push_back(x + y);
    if(activeStrategy == INACTIVE) {
      /**
       * Find new strategy
       */
      if(y < 0){
      /**
       * Ignore Strategy, maybe error in strategy
       */

      }
      else if( (x + y) >= tempStategyValue  && (x + y) > 0 ) {
	tempStrategy = i;
	tempStategyValue = x + y;
	
      }
    }
    else {
      /**
       * Keep active strategy
       * Unless emergency
       */
      if(y < 0){
      /**
       * Ignore Strategy, maybe error in strategy
       */

      }
      else if( (x + y) >= (activeStrategyValue + 100) && (x + y)>0) {
	tempStrategyEmergency = i;

      }
      else 
	tempStrategy = activeStrategy;

    }
  }

    /** ********************** *
     * Set the active strategy *
     * *********************** *
     */
    
    if(activeStrategy == INACTIVE) {
      /**
       * New Strategy
       */
      activeStrategy = tempStrategy;
      if(tempStrategy != INACTIVE){
        activeStrategyValue = tempStrategyValueVector.at(tempStrategy);
        _observers.at(activeStrategy)->SetAgentActive( ACTIVE );
      }
      tempStrategyValueVector.clear();
      
    }
    else if((tempStrategy==activeStrategy)&&(tempStrategyEmergencyValue < activeStrategyValue+100)){
     /**
      * Keep existing Strategy
      */
      activeStrategyValue = tempStrategyValueVector.at(activeStrategy);
      tempStrategyValueVector.clear();

    }
    else if ( tempStrategyEmergencyValue > ( activeStrategyValue + 100 ) ) {
     /**
      * Do not keep existing strategy,
      * or act on emergency.
      * Inform the active strategy
      */
      _observers.at(activeStrategy)->StrategyInterrupted();
      _observers.at(activeStrategy)->SetAgentActive( INTERRUPTED );
      interruptedStrategyes.push_back(activeStrategy);
      activeStrategy = tempStrategyEmergency;

    }

#ifdef DEBUG_ACTION_PRINT
      cout << "activStrategy =  " << activeStrategy << endl;

#endif

      tempStrategyValueVector.clear();
#ifdef DEBUG_ACTION_PRINT
      cout << "**** 2 ACTIVE STRATERY ***** = " << activeStrategy << endl;
#endif
}

int PlansysObservation::PerformInterruptStrategy( interrupt_controll_struct *ai_interrupt_controll_struct ) {
  int i;
  for( i = 0; i < _observers.size(); i++ ) {
    (_observers.at(i))->GetInterruptAction(ai_interrupt_controll_struct);
  }
  
  return 0;
}

int PlansysObservation::PerformStrategy( controll_struct *ai_controll_struct ) {
  if(activeStrategy < 0 || activeStrategy > _observers.size()){
   // Do nothing
#ifdef DEBUG_ACTION_PRINT
    cout << " " << endl;
    cout << " " << endl;
    cout << " " << endl;
    cout << " " << endl;
    cout << "activeStrategy out of range " << activeStrategy << endl;
    cout << " " << endl;
    cout << " " << endl;
    cout << " " << endl;
#endif
    return -1;
  }
  else {
    /**
     * Get action
     */
    (_observers.at(activeStrategy))->GetAction(ai_controll_struct);

    return 0;
  }
}

int PlansysObservation::ChoseStrategy() {


  return 0;
}

ai_waypoint PlansysObservation::GetPosObservation() {

  return ai_waypoint_3d;
}

ai_waypoint PlansysObservation::GetWaypointReachedObservation() {
  return last_ai_waypoint_3d_reached;
}

cv_observation PlansysObservation::GetCvObservation() {
  return my_cv_observation;
}

cv_observation PlansysObservation::GetCvColorObservation() {
  return color_cv_observation;
}

time_message PlansysObservation::GetTimeObservation() {
  return main_time;
}

proxy_message PlansysObservation::GetProxyMessage() {
 return my_proxy_message; 
}

collision_detection_message PlansysObservation::GetCollisionMessage() {
  return collision_message;
}

int PlansysObservation::ClearTheActiveStrategy() {
  int i;
  for( i = 0; i < _observers.size(); i++ ) 
    _observers.at(i)->SetAgentActive( INACTIVE );
  activeStrategy = INACTIVE; 
  activeStrategyValue = 0;
  return 0;
}

int PlansysObservation::GetAvtiveStrategy() {
  return activeStrategy;
}
