/***************************************************************************
 *   Copyright (C) 2007 by Gunnar Kjemphol   *
 *   gunnar@kjemphol.no                                                    *
 *   Eurobot 2007                                                          *
 *   The Eurobot AI (Plansys) program                                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif 


#include <iostream>
#include <cstdlib>
#include <sys/time.h>
#include <math.h>

#include "sendRecieve.h"
#include "MessageHandler.h"
#include "posix_message_queue.h"
#include "posix_message_config.h"
#include "observations.h"
#include "Observer.h"
#include "PlansysObservation.h"
#include "BottleBoxSort.h"
#include "BatterySort.h"
#include "MotorAi.h"
#include "FirstStrategy.h"
#include "waypoint.h"
//#include "WorldModel.h"

#include "mainplansys.h"
#include "init.h"
#include "AStarUse.h"

int teamColor = 0;
int recievedTeamColor = FALSE;

int collisionDetect = 0;
int askForCvObservationFlag = FALSE;


int recievedGameStartMessage = FALSE;
int gameResetMessageTimer = TRUE;
int gameResetMessageMain = TRUE;
int gameFirstStart = TRUE;
int newPickUpSucessFlag = FALSE;

pthread_mutex_t mutex_game_reset_timer;
pthread_mutex_t mutex_game_reset_main;
pthread_mutex_t mutex_recieved_game_start_message;


using namespace std;
/**
 * Define a PlansysObservation object
 */
PlansysObservation plansysObs;

/**
 * **************************************************************************** *
 * Internal Controll functions                                                  *
 * **************************************************************************** *
 */



extern "C" void* timer_thread(void *arg){
  int m;
  /**
   * Timer, turn off actuators after (GAME_TIME) sec
   */
  MessageHandler *proxysysMessageHandler;
  proxysysMessageHandler = (MessageHandler *)arg;
  
  time_message main_time_struct, *main_time;
  main_time = &main_time_struct;
  main_time->timeLeft = -1;
  main_time->gameTime = GAME_TIME;

  pmq_message pos_mes;
  pos_mes.to = -1;
  pos_mes.from = -1;
  pos_mes.major = -1;
  pos_mes.minor = -1;
  pos_mes.priority = -1;
  pos_mes.length = sizeof(int);


  int i = 0;
  timeval tim;
  double t1, t2, t;
  gettimeofday(&tim,NULL);
  t1 = tim.tv_sec;
  while(1){

    if(gameResetMessageTimer == TRUE){
      usleep(50000);
      while(recievedTeamColor == FALSE){
	usleep(1000);
      }
      usleep(50000);
      while(recievedGameStartMessage == FALSE){
	usleep(1000);
      }
      i = 0;
      gettimeofday(&tim,NULL);
      t1 = tim.tv_sec;
      proxysysMessageHandler->sendCommand( MINOR_CMD_SYSTEM_GO, PRIORITY_HIGHEST );
    }
    usleep(1000);
    gettimeofday(&tim,NULL);
    t2 = tim.tv_sec;
    t = t2 - t1;
    main_time->timeLeft = (main_time->gameTime - (int) t);
#ifdef DEBUG_TIME
//    for(i = 0; i < 50; i++)
//    cout << "timeLeft " << main_time->timeLeft << endl;
#endif

    /**
     * Update time on the Agents
     */
    i++;
    if(i > 10){
      pos_mes.length = main_time->timeLeft;
      plansysObs.NewObservation(pos_mes, GAME_TIME_MESSAGE);
      i = 0;
    }

    if(t2 > t1 + GAME_TIME){
      gameResetMessageTimer = TRUE;
      gameResetMessageMain = TRUE;
      recievedTeamColor = FALSE;
      recievedGameStartMessage = FALSE;
      cout << "Time is up " << endl;
      /**
       * Send Command turn off, this command will turn off all actuators
       */
      proxysysMessageHandler->sendCommand( MINOR_CMD_SYSTEM_STOP, PRIORITY_HIGHEST );
      for (m = 0; m < 10; m++) {
      cout << "Time is up " << endl;
      cout << "Time is up " << endl;
      }

      /**
       * Set new time when reset
       */
      while( gameResetMessageTimer == FALSE ) {
	i = 0;
	gettimeofday(&tim,NULL);
        t1 = tim.tv_sec;
      }
      break;
    }
  }
}


/**
 *  *************************************************************************** *
 *  Init
 * ************************
 */

int initGame() {
  teamColor = 0;
  recievedTeamColor = FALSE;

  recievedGameStartMessage = FALSE;
  gameResetMessageTimer = TRUE;
  gameResetMessageMain = TRUE;
  gameFirstStart = TRUE;
  return 0;
}


/**
 * **************************************************************************** *
 * Main function                                                                *
 * **************************************************************************** *
 */

int main(int argc, char *argv[]) {

  signal(SIGHUP, catch_sighup);

  int x;
  int i = 0;
  int count = 0;
  timeval cv_timer;
  bool stopRobot = false;

  double cv_t1, cv_t2;
  gettimeofday(&cv_timer,NULL);
  cv_t1 = cv_timer.tv_sec;
  cv_t2 = cv_t1;

  int cv_counter = 4; // Use for switching between BOTTLE_BOX 
                      // and BATTERY observation

  bool last_waypoint_sendt_delay = FALSE;

  initGame();

  /**
   * Init A* World Grid
   */
  InitMap( 12, 8 );
  
  // Define cv_observation struct
  cv_observation my_cv_observation;
  clear_cv_observation(&my_cv_observation);

  // Define and clear ai_waypoint
  ai_waypoint my_ai_waypoint;
  clear_robot_state_3d( &my_ai_waypoint );
  
  // Define and clear speed limit
  speed_limit max_speed_limit;
  max_speed_limit.u_b = MAX_SPEED_MPS;
  max_speed_limit.r_b = 10.0;

#ifdef DEBUG_INIT_PRINT
  cout << "Robot AI" << endl;
#endif

  /**
   * Define interrupt controll struct
   */
  interrupt_controll_struct ai_interrupt_controll_struct;
  clear_interrupt_controll_struct( &ai_interrupt_controll_struct );

  /**
   * Define controll struct
   */
  controll_struct ai_controll_struct;
  clear_controll_struct( &ai_controll_struct );

  /**
   * Declare Message Handlers
   */
  MessageHandler cvsysMessageHandler(PLANSYS_ID, PMQ_PLANSYS, PMQ_CVSYS, false);
  MessageHandler cvsysColorMessageHandler(PLANSYS_ID, PMQ_PLANSYS, PMQ_CVSYS_COLOR, false);
  MessageHandler proxysysMessageHandler(PLANSYS_ID, PMQ_PLANSYS, PMQ_PROXY, false);
  MessageHandler navsysMessageHandler(PLANSYS_ID, PMQ_PLANSYS, PMQ_NAVSYS, false);

  /**
   * Declare sub Message Handlers with interrupt
   */

  /**
   * Declare sub Message Handlers without interrupt
   */
  navsysMessageHandler.addSubMessageHandler(recievePositionUpdate, MAJOR_STRUCT, MINOR_NAVSYS_POSITION);

  navsysMessageHandler.addSubMessageHandler(recieveWaypointReached, MAJOR_STRUCT, MINOR_NAVSYS_REACHED_WAYPOINT);
  
  cvsysMessageHandler.addSubMessageHandler(recieveCvObservaton, MAJOR_STRUCT, MINOR_CV_OBSERVATION);
  
  cvsysColorMessageHandler.addSubMessageHandler(recieveCvColorObservaton, MAJOR_STRUCT, MINOR_CV_COLOR_OBSERVATION);
  
  proxysysMessageHandler.addSubMessageHandler(recieveProxyMessage, MAJOR_STRUCT, MINOR_PROXY_MESSAGE);
  proxysysMessageHandler.addSubMessageHandler(recieveCmdProxyMessage, MAJOR_COMMAND, MINOR_CMD_SYSTEM_GO);
  proxysysMessageHandler.addSubMessageHandler(recieveCmdProxyMessage, MAJOR_COMMAND, MINOR_CMD_SYSTEM_INITIALIZE);
  proxysysMessageHandler.addSubMessageHandler(recieveCmdTeamColor, MAJOR_COMMAND, MINOR_CMD_SET_TEAM_COLOR_RED);
  proxysysMessageHandler.addSubMessageHandler(recieveCmdTeamColor, MAJOR_COMMAND, MINOR_CMD_SET_TEAM_COLOR_BLUE);
  proxysysMessageHandler.addSubMessageHandler(recieveCmdBatterySort, MAJOR_COMMAND, MINOR_CMD_BATTERY_RELEASED);
  proxysysMessageHandler.addSubMessageHandler(recieveCmdBottleBoxSort, MAJOR_COMMAND, MINOR_CMD_BOTTLE_BOX_RELEASED);
  proxysysMessageHandler.addSubMessageHandler(recieveCmdBottleBoxSort, MAJOR_COMMAND, MINOR_CMD_PICK_UP_BOTTLE_SUCCESS);
  proxysysMessageHandler.addSubMessageHandler(recieveCmdBottleBoxSort, MAJOR_COMMAND, MINOR_CMD_PICK_UP_BOX_SUCCESS);
  proxysysMessageHandler.addSubMessageHandler(recieveCmdBottleBoxSort, MAJOR_COMMAND, MINOR_CMD_PICK_UP_FAILURE);
  proxysysMessageHandler.addSubMessageHandler(recieveCollisionDetectionStruct, MAJOR_STRUCT, MINOR_COLLISION_DETECTED);


  /**
   * Wait for the teamColor message
   */
  waitForTeamColorMessage( &proxysysMessageHandler, &navsysMessageHandler );

  gameResetMessageTimer = FALSE;
  SendInitPosition( &navsysMessageHandler );

  /**
   * Wait for the game Start message
   */
    while( recievedGameStartMessage == FALSE){ 
      cout << "Wait for game start message " << endl;
      usleep(100000);
      navsysMessageHandler.handleAllMessagesNow();

    }

    /**
     * Send Go message to PROXY
     * PROXY Sends message to navsys and BottleBoXSort hardware
     * 
     */
    proxysysMessageHandler.sendCommand( MINOR_CMD_SYSTEM_GO, PRIORITY_HIGHEST );

 /**
  * Ask for CV battery_bin position
  * BATTERY_BIN
  */
    my_cv_observation.type = BATTERY_BIN;
    my_cv_observation.teamColor = (short)teamColor;
    cvsysMessageHandler.sendMessage( (char *)&my_cv_observation, (int)sizeof(cv_observation), MINOR_CV_OBSERVATION, PRIORITY_VERY_HIGH );


 /**
  * Init and Start the timer thread
  */
  pthread_t thread1;
  pthread_create(&thread1, NULL, timer_thread, &proxysysMessageHandler);

/**
 * Declare new AI Agents.
 */
  MotorAi motorAi( &plansysObs ,teamColor );
  FirstStrategy firstStrategy( &plansysObs ,teamColor );
  //  BatterySort batterySort( &plansysObs ,teamColor );
  BottleBoxSort bottleBoxSort( &plansysObs ,teamColor );
  //  EmptyStrategy emptyStrategy( &plansysObs ,teamColor );

  gameFirstStart = FALSE;
  gameResetMessageMain = FALSE;

/**
  * ******************************************************************* *
  *                MAIN WHILE LOOP                                      *
  * ******************************************************************* *
  */
  while(1){

    if(newPickUpSucessFlag == TRUE) {
     /**
      * New pickup sucsess ask cvColorSystem for the right type
      */
     for(i = 0; i < 100; i++){
    //   cout << " newPickUpSucessFlag == TRUE " << endl; 
      usleep(1000);
     }
     clear_cv_observation(&my_cv_observation);
      my_cv_observation.type = BOTTLE_BOX;
      cvsysColorMessageHandler.sendMessage( (char *)&my_cv_observation, (int)sizeof(cv_observation), MINOR_CV_OBSERVATION, PRIORITY_VERY_HIGH );
      newPickUpSucessFlag = FALSE;
    }

    if( recievedTeamColor == TRUE){ 
      cout << "RecieveteamColor " << endl;
    /**
     * Set the team color LED
     */
      if( teamColor == TEAM_COLOR_RED )
	proxysysMessageHandler.sendCommand(MINOR_SET_LED_TEAM_COLOR_RED ,PRIORITY_HIGH);
      else if( teamColor == TEAM_COLOR_BLUE )
	proxysysMessageHandler.sendCommand(MINOR_SET_LED_TEAM_COLOR_BLUE ,PRIORITY_HIGH);
      recievedTeamColor = FALSE;

    }

   if( gameResetMessageMain == TRUE ) {
    cout << "gameResetMessageMain == TRUE" << endl;
    /**
     * Flash the sensor board LEDS
     */
    flashTheSensorBoardLeds( &proxysysMessageHandler );

  /**
   * Wait for the teamColor message
   */
    recievedTeamColor = FALSE;

   /**
    * Stop Robot
    */
    while(1){
      max_speed_limit.u_b = 0x00;
      max_speed_limit.r_b = 0x00;
      
      navsysMessageHandler.sendMessage( (char *)&max_speed_limit, (int)sizeof(speed_limit), MINOR_NAVSYS_SET_SPEED_LIMIT, PRIORITY_HIGHEST );
  
      usleep(300000);
    }

    waitForTeamColorMessage( &proxysysMessageHandler, &navsysMessageHandler );

  /**
   * Wait for the game Start message
   */
    recievedGameStartMessage = FALSE;
    while( recievedGameStartMessage == FALSE){ 
      cout << "Wait for game start message " << endl;
      usleep(1000000);
      navsysMessageHandler.handleAllMessagesNow();

    }
    /**
     * Reset timer
     */
    gameResetMessageMain = FALSE;
  } //if( gameResetMessageMain == TRUE )

    /**
     * Ask for position
     */
#ifdef DEBUG_CMD_PRINT
    cout << "PositionRequest" << endl;
#endif
    navsysMessageHandler.sendCommand( MINOR_CMD_REQUEST_POSITION, PRIORITY_HIGH );
    /**
     * Ask for CV observation only if activeStrategy = INACTIVE
     * Ask for CV observation ask for battery for every fift call
     */
  if(askForCvObservationFlag == TRUE){
    if( plansysObs.GetAvtiveStrategy() == INACTIVE ) {
      gettimeofday(&cv_timer,NULL);
      cv_t1 = cv_timer.tv_sec;
      cv_t2 = cv_t1;
      askForCvObservation( &cvsysMessageHandler, &cv_counter );
      usleep(400000);
    }
  }

#ifdef DEBUG_CMD_PRINT
    cout << "Sleep" << endl;
#endif
    usleep(400000);

  /**
   * Check the posix queues for new messages
   */
    navsysMessageHandler.handleAllMessagesNow();

    /**
     * Get the priority form the AI Agents
     */
#ifdef DEBUG_CMD_PRINT
    cout << "plansysObs.GetAgentPriority( ); "<< endl;
#endif
    plansysObs.GetAgentPriority( );


/**
 * *************************************************************
 *            AI INTERRUPT CONTROLL STRUCT                     *
 * *************************************************************
 */


plansysObs.PerformInterruptStrategy( &ai_interrupt_controll_struct );
if( ai_interrupt_controll_struct.stopRobot == true ){
  max_speed_limit.u_b = 0.0;
  max_speed_limit.r_b = 10.0;
  navsysMessageHandler.sendMessage( (char *)&max_speed_limit, (int)sizeof(speed_limit), MINOR_NAVSYS_SET_SPEED_LIMIT, PRIORITY_HIGHEST );
  for(x = 0; x < 10; x++){
    usleep(50000);
    cout << "COLLISION DETECTION " << endl;
  }
  stopRobot = true;
  // Clear
  ai_interrupt_controll_struct.stopRobot = false;

}

#ifdef DEBUG_CMD_PRINT
cout << "max_speed_limit.u_b mps = " << max_speed_limit.u_b << endl;
#endif

if( ai_interrupt_controll_struct._interrupt_ai_waypoint.size() > 0 ){
    max_speed_limit.u_b = MAX_SPEED_MPS;
    max_speed_limit.r_b = 10.0;

  for(count = 0; count < ai_interrupt_controll_struct._interrupt_ai_waypoint.size(); count++){
    navsysMessageHandler.sendMessage( (char *)&ai_interrupt_controll_struct._interrupt_ai_waypoint.at(count), (int)sizeof(ai_waypoint), MINOR_NAVSYS_RELATIVE_INTERRUPT_WAYPOINT, PRIORITY_HIGH );
  }
  ai_interrupt_controll_struct._interrupt_ai_waypoint.clear();
  usleep(500000);
  navsysMessageHandler.sendMessage( (char *)&max_speed_limit, (int)sizeof(speed_limit), MINOR_NAVSYS_SET_SPEED_LIMIT, PRIORITY_HIGHEST );

}

#ifdef PRINT_DEBUG
  cout << "ai_controll_struct _ai_waypoint.size "<< ai_controll_struct._ai_waypoint.size() << endl;
    
  cout << "ai_controll_struct _ai_waypoint_rel_robot.size "<< ai_controll_struct._ai_waypoint_rel_robot.size() << endl;
#endif

  plansysObs.PerformStrategy( &ai_controll_struct );

/**
 * *************************************************************
 *                     AI CONTROLL STRUCT                      *
 * *************************************************************
 */

    /**
     * AI CONTROLL STRUCT, this struct is edited by the ACTIVE
     * strategy.
     */
#ifdef PRINT_DEBUG
    cout << "ai_controll_struct _ai_waypoint.size "<< ai_controll_struct._ai_waypoint.size() << endl;
    cout << "ai_controll_struct _ai_waypoint_rel_robot.size "<< ai_controll_struct._ai_waypoint_rel_robot.size() << endl;
#endif
  /**
   * Set max speed
   */
    if(ai_controll_struct.setMaxSpeed == true) {
      max_speed_limit.u_b = ai_controll_struct.maxSpeed_mps;
      max_speed_limit.r_b = 10.0;
      navsysMessageHandler.sendMessage( (char *)&max_speed_limit, (int)sizeof(speed_limit), MINOR_NAVSYS_SET_SPEED_LIMIT, PRIORITY_HIGHEST );
       ai_controll_struct.setMaxSpeed == false;
    }
  /**
    * Send message to clear waypoints
    */
  if(ai_controll_struct.clearWaypoints == true){
    navsysMessageHandler.sendCommand( MINOR_CMD_CLEAR_WAYPOINTS, PRIORITY_HIGHEST);

  }
  /**
   * The Active strategy is finished
   */
  if( ai_controll_struct.clearTheActiveStrategy == true ) {
    plansysObs.ClearTheActiveStrategy();
    ai_controll_struct.clearTheActiveStrategy = false;
    last_waypoint_sendt_delay = TRUE;
#ifdef DEBUG_CMD_PRINT
    cout << "ai_controll_struct.clearTheActiveStrategy = true;" << endl;
#endif
  }

  /**
   * Send The Command from the command struct
   */
  if(ai_controll_struct.command.size() > 0) {
    for( count = 0; count < ai_controll_struct.command.size(); count++ ) {
#ifdef PRINT_DEBUG
      cout << "ai_controll_struct.command.st(count)" << ai_controll_struct.command.at(count) << endl;
#endif
      proxysysMessageHandler.sendCommand( ai_controll_struct.command.at(count), PRIORITY_HIGHEST );

    }
    ai_controll_struct.command.clear();
    sleep(1);
  }


  if(ai_controll_struct._ai_waypoint_rel_robot.size()){
    for( count = 0; count < ai_controll_struct._ai_waypoint_rel_robot.size(); count++ ) {
#ifdef PRINT_DEBUG
      cout << "Send waypoint rel robot" << endl;
#endif
      /**
       * Send waypoints relative to robot
       */
      navsysMessageHandler.sendMessage( (char *)&ai_controll_struct._ai_waypoint_rel_robot.at(count), (int)sizeof(ai_waypoint), MINOR_NAVSYS_SET_WAYPOINT_ROBOT_RELATIVE, PRIORITY_NORMAL );
      if(last_waypoint_sendt_delay == TRUE) {
        for (i = 0; i < 4; i++){
          cout << "main: sleep 4" << endl;
          sleep(1);
        }
        last_waypoint_sendt_delay = FALSE;
      }
    }
   /**
    * Clear the waypoint_relative struct
    */
    ai_controll_struct._ai_waypoint_rel_robot.clear();
  }


  if(ai_controll_struct._ai_waypoint.size()){
    for( count = 0; count < ai_controll_struct._ai_waypoint.size(); count++ ) {

      /**
      * Send waypoints
      */
      navsysMessageHandler.sendMessage( (char *)&ai_controll_struct._ai_waypoint.at(count), (int)sizeof(ai_waypoint), MINOR_NAVSYS_SET_WAYPOINT, PRIORITY_NORMAL );
    }
    /**
    * Clear the waypoint struct
    */
    ai_controll_struct._ai_waypoint.clear();
  }
    clear_controll_struct( &ai_controll_struct );
}

  return 0;
}

void catch_sighup(int signum){
  signal(SIGHUP, catch_sighup);
}

